/***
 *** HelperDialog.cpp
 ***
 *** This is the source code for the helper dialog used by TTCN based test suites.  This source file should not be included
 *** with the implicit send project distributed through PTS.
 ***
 *** ----------
 ***
 *** Copyright 2012, Bluetooth SIG, Inc.
 ***
 *** ----------
 ***
 *** Bradford Ayres
 *** Bluetooth SIG, Inc.
 ***/


#define	WINVER	0x0501	// Windows XP

#include	<afxwin.h>

#include	"implicitsendapi.h"
#include	"helperdialog.h"
#include	"mycstatic.h"
#include	"resource.h"
#include    "MSXMLWrapper.h"

#define REGISTRY_PATH _T("software\\bluetooth sig\\bluetooth PTS 4")
extern std::string	m_strBdAddr;	// Bluetooth Address apend to the caption

/**
 * \class ThreadSafeObject
 * \brief This object manages a value in a thread safe manner.
 */
template <class T>
class ThreadSafeObject {

public :

	/**
	 * Constructor
	 */
	ThreadSafeObject( const T& value ) : m_Value(value){
		InitializeCriticalSection( &m_CriticalSection );
	}

	/**
	 * Destructor
	 */
	virtual ~ThreadSafeObject( void )
	{
		DeleteCriticalSection( &m_CriticalSection );
	}

	/**
	 * Sets the current value.
	 * @param The new value to set.
	 */
	void SetValue( const T &value ) {
		EnterCriticalSection( &m_CriticalSection );
		m_Value = value;
		LeaveCriticalSection( &m_CriticalSection );
	}

	/**
	 * Gets the current value
	 * @return The current value.
	 */
	T GetValue( void ) {
		EnterCriticalSection( &m_CriticalSection );
		std::string value = m_Value;
		LeaveCriticalSection( &m_CriticalSection );
		return value;
	}

private :

	/**
	 * The current value.
	 */
	T m_Value;

	/**
	 * The critical section used to synchronize access.
	 */
	CRITICAL_SECTION m_CriticalSection;
};

/**
 * Static string used to communicate what command was selected.
 */
static ThreadSafeObject<std::string> s_SelectedCommand("");


/**
 * Class that helps manage the helper dialog.
 */
class CHelperThread {

public :

	/**
	 * Constructor
	 */
	CHelperThread( void ) {
		m_hEvents[0] = CreateEvent( NULL, TRUE, FALSE, "CommandSelected" );
	}

	/**
	 * Destructor
	 */
	virtual ~CHelperThread( void ) {
		CloseHandle( m_hEvents[0] );

		// add additional thread shutdown code here 
		if ( m_hEvents[1] )
		{
			if ( !TerminateThread( m_hEvents[1], 0 ) )
			{
				// Failed to terminate thread
			}
			else
			{
				// Wait for thread to terminate
				WaitForSingleObject( m_hEvents[1], INFINITE );
			}

			// Close the thread
			CloseHandle( m_hEvents[1] );

			// Clear handle
			m_hEvents[1] = NULL;
		}
	}

	/**
	 * ThreadProc
	 */
	static DWORD WINAPI helperThread(LPVOID arg)
	{
		try {
			AFX_MANAGE_STATE(AfxGetStaticModuleState());
			CHelperThread* pThis = reinterpret_cast<CHelperThread*>(arg);
			CHelperDialog helperDialog( pThis->m_XmlString, pThis->m_hEvents );
			helperDialog.DoModal();
		}
		catch( std::exception &e ) {
			// Failed to load helper
			MessageBox( NULL, e.what(), "Helper Error", MB_OK );
			return 0;
		}

		return 0;
	}

	/**
	 * Method waits for a command to be selected, or for the dialog to be closed.
	 * @param xmlString If the dialog is not active, this string will be used to initialize the dialog.
	 * @return An empty string if the dialog was closed, otherwise the name of the command that was selected.
	 */
	std::string WaitForEvent( const std::string &xmlString )
	{
		if ( !m_hEvents[1] ) {
			// Initialize thread now
			DWORD threadId = 0;
			m_XmlString = xmlString;
			m_hEvents[1] = CreateThread( NULL, 0, helperThread, this, 0, &threadId );
		}

		// Reset Events
		ResetEvent( m_hEvents[0] );

		// Wait for an event to trigger
		WaitForMultipleObjects( 2, m_hEvents, FALSE, INFINITE );

		// Check if the dialog thread was closed
		DWORD exitCode = 0;
		if ( GetExitCodeThread( m_hEvents[1], &exitCode ) )
		{
			if ( exitCode != STILL_ACTIVE )
			{
				// Dialog thread closed, clean up thread info now
				CloseHandle( m_hEvents[1] );
				m_hEvents[1] = NULL;
				return "";
			}
			else
			{
				// Command was executed
				return s_SelectedCommand.GetValue();
			}
		}
		else
		{
			throw std::exception( "Error with helper thread." );
		}
	}

private :

	/**
	 * Events to wait for.  HANDLE 0 is used to signal a command was selected, HANDLE 1 is tied to the thread.
	 */
	HANDLE m_hEvents[2];
	
	/**
	 * The Xml String used to setup the dialog.
	 */
	std::string m_XmlString;
};


/**
 * Helper dialog interface used by the DLL, displayes the Helper MMI and waits for an action
 */
std::string RunHelperDialog( const std::string &xmlString, const std::string &strBdAddr )
{
	m_strBdAddr = strBdAddr;
	static CHelperThread s_HelperThread;
	return s_HelperThread.WaitForEvent( xmlString );
}


/**
 * MFC Message Map
 */
BEGIN_MESSAGE_MAP(CHelperDialog,CDialog)
ON_LBN_DBLCLK(IDC_HELPER_COMMANDS, OnCommandListDoubleClick)
ON_LBN_SELCHANGE(IDC_HELPER_COMMANDS, OnCommandSelectionChanged)
ON_CBN_SELCHANGE(IDC_HELPER_FILTER, &CHelperDialog::OnCbnSelchangeHelperFilter)
ON_WM_CLOSE()
END_MESSAGE_MAP()


 CHelperDialog::CHelperDialog( const std::string &xmlString, HANDLE *hEvents ) :
	CDialog(IDD_IMPLICITSENDHELPER), m_InitializationXmlString( xmlString )
{
	m_hEvents[0] = hEvents[0];
}

CHelperDialog::~CHelperDialog()
{
	// event is owned by the CHelperThread class, don't close it
}

void CHelperDialog::DoDataExchange(
	CDataExchange	*pDx)
{
	// Do default
	CDialog::DoDataExchange(pDx);

	// Connect to controls
	DDX_Control(pDx,IDC_HELPER_FILTER,m_FilterComboBox);
	DDX_Control(pDx,IDC_HELPER_COMMANDS,m_CommandListBox);
	DDX_Control(pDx,IDC_HELPER_COMMAND_INFO,m_CommandDescription);
	//DDX_Control(pDx,IDC_HELPER_INFO,m_HelperDescription);
}

void CHelperDialog::SetActiveFilter( const std::string &filterName )
{
	// Set the active filter by name
	if ( m_Filters.find( filterName ) != m_Filters.end() ) 
	{
		m_Filters.find( filterName )->second.SetActiveFilter( m_CommandListBox );
	}
	else if ( m_Filters.begin() != m_Filters.end() )
	{
		// Can't find the requested name, just set the first filter in the list
		m_Filters.begin()->second.SetActiveFilter( m_CommandListBox );
		m_FilterComboBox.SelectString( 0, m_Filters.begin()->first.c_str() );
	}
	else
	{
		// Nothing to set, do nothing
	}

	// clear the command description box
	m_CommandDescription.SetWindowText( "" );
}

void CHelperDialog::InitializeCommand( const std::string &helperName, MSXML2::IXMLDOMNodePtr commandNode )
{
	// Load the command from the given XML node
	std::string name = "", role = "", description = "";
	
	// Get name
	MSXML2::IXMLDOMNodeListPtr nodeList = NULL;
	MSXML2::IXMLDOMNodePtr node = commandNode->selectSingleNode( "@name" );
	if ( !node )
	{
		// name must exist
		throw std::exception( "Each <command> element must have a name attribute" );
	}
	else
	{
		name = node->Gettext();
	}

	// Get role
	node = commandNode->selectSingleNode( "@role" );
	if ( node )
	{
		role = node->Gettext();
	}

	// Get description
	node = commandNode->selectSingleNode( "description" );
	if ( node )
	{
		description = node->Gettext();
	}

	// create command
	CHelperDialog::Command *pCommand = new CHelperDialog::Command( helperName, name, role, description );

	if ( !pCommand )
	{
		// An error occured
		throw std::exception( "Failed to allocate command object." );
	}
	else
	{
		// Add command to default filters
		m_Filters["All Commands"].AddCommand( pCommand );
		m_Filters[helperName].AddCommand( pCommand );

		// Look for filter attribute
		node = commandNode->selectSingleNode( "@filter" );
		if ( node )
		{
			m_Filters[std::string(node->Gettext())].AddCommand( pCommand );
		}

		// Look for filter elements
		nodeList = commandNode->selectNodes( "filter" );
		for ( int i = 0; i < nodeList->length; i++ )
		{
			// Add command to the specified filter
			std::string temp = nodeList->item[i]->text;
			m_Filters[temp].AddCommand( pCommand );
		}
	}
}

std::string GetCurrentModulePath( const char* pathToRemove )
{
	char	installationPath[_MAX_PATH];

	// Try to get the implicit send dll path
	HMODULE dllHandle = NULL;
	if (GetModuleHandleEx(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS |
		GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT, (LPCTSTR)GetCurrentModulePath, &dllHandle))
	{
		if (GetModuleFileName(dllHandle, installationPath, _countof(installationPath)))
		{
			char* pos = NULL;
if (pathToRemove)
			{
				pos = strstr(installationPath, pathToRemove);
			}
			else
			{
				// Nothing specified for remomval
			}
			
			if (pos)
			{
				// Truncate the string based on the pathToRemove
				pos[1] = '\0';
			}
			else
			{
				// Nothing to remove
			}
			
			return installationPath;
		}
		else
		{
			// Failed to get module file name
		}
	}
	else
	{
		// Failed to get module handle
	}

	return "";
}

void CHelperDialog::InitializeHelpers( std::string &helperList )
{
	// Load the master helper XML
	CMSXMLWrapper xmlWrapper;
	MSXML2::IXMLDOMNodeListPtr commandList = NULL;
	MSXML2::IXMLDOMNodePtr command = NULL;

	// Get Installation Path
	std::string	installationPath = GetCurrentModulePath("\\bin\\implicit_send3.dll");
	if (installationPath.empty())
	{
		// Failed to get installation path
	}
	else
	{
		MSXML2::IXMLDOMDocumentPtr root = xmlWrapper.LoadFromFile(installationPath + "data\\helper\\helper.xml");
		if (!root)
		{
			// Failed to load helper XML file
			throw std::exception("Failed to load helper.xml");
		}

		std::string::size_type right = helperList.length();
		std::string::size_type left = std::string::npos;

		// Loads the helper
		std::string nextHelper = "";
		while (!helperList.empty())
		{
			// get the name of the last helper in the list
			left = helperList.find_last_of(";");
			if (left == std::string::npos)
			{
				// the current string is the name of the helper
				nextHelper = helperList;
				helperList = "";
			}
			else
			{
				// pull out the helper name and update the helper list
				nextHelper = helperList.substr(left + 1, right - left);
				helperList = helperList.substr(0, left);
			}

			// Load all commands from the associated helper
			std::string path = (std::string("//helper[@name='") + nextHelper + std::string("']/command"));
			commandList = root->selectNodes(path.c_str());
			for (int i = 0; i < commandList->length; i++)
			{
				// parse the command and add it to the helpers list of known commands
				InitializeCommand(nextHelper, commandList->item[i]);
			}
		}

		// Initialize combo box
		std::map<std::string, CHelperDialog::Filter>::iterator itFilter = m_Filters.begin();
		while (itFilter != m_Filters.end())
		{
			std::string temp = itFilter->first.c_str();
			m_FilterComboBox.AddString(itFilter->first.c_str());
			itFilter++;
		}

		// Set the default filter to the last helper loaded
		m_FilterComboBox.SelectString(0, nextHelper.c_str());

		// Set Active Filter
		SetActiveFilter(nextHelper.c_str());
	}
}

BOOL CHelperDialog::OnInitDialog()
{
	// Perform default actions first
	CDialog::OnInitDialog();

	// Set icon
	SetIcon(AfxGetApp()->LoadIcon(IDI_BLUETOOTH_LOGO),TRUE);

	//try {
		// Clear text fields
		m_CommandDescription.SetWindowText( "" );
		//m_HelperDescription.SetWindowText( "" );

		// Use XML string to get description, and helpers to load
		std::string helperNames = InitializeFromXMLString();

		// Use helper names to initialize each helper
		InitializeHelpers( helperNames );

		return TRUE;
	//}
	//catch( ... )
	//{
	//	// Failed to initialize
	//	return FALSE;
	//}
}

std::string CHelperDialog::InitializeFromXMLString( void )
{
	CMSXMLWrapper xmlWrapper;
	MSXML2::IXMLDOMNodePtr element = NULL;
	MSXML2::IXMLDOMNodePtr root = xmlWrapper.LoadFromString( m_InitializationXmlString );
	
	if ( !root )
	{
		return FALSE;
	}

	// Set Helper Description Text
	std::string descriptionText = "";
	element = root->selectSingleNode( "//description" );
	if ( element ) {
		descriptionText = element->Gettext();
	}

	// TODO: Change title to allow for MMI and Helper active at the same time
	// TODO: Will need to create another TSO_close_mmi to close the helper
	//SetWindowText( (std::string( "PTS Helper - " ) + descriptionText).c_str() );
	if (m_strBdAddr != "")
	{
		std::string strTitle = "User Action Required - " + m_strBdAddr;
		SetWindowText( strTitle.c_str() );
	}
	else
		SetWindowText( "User Action Required" );

	// Load helpers
	std::string helpers = "";
	element = root->selectSingleNode( "//helpers" );
	if ( !element )
	{
		// XML must have a helper element
		return FALSE;
	}
	else
	{
		// get the list of helpers
		helpers = element->Gettext();
		if ( helpers.empty() )
		{
			// helper element must have text
			return FALSE;
		}
	}

	// return the list of helper names
	return helpers;
}

void CHelperDialog::DisplayCommandDescription( void )
{
	// update the command description field
	int index = m_CommandListBox.GetCurSel();
	if ( index == -1 )
	{
		// No selection
		m_CommandDescription.SetWindowTextA( "" );
	}
	else
	{
		// Set the associated description
		CHelperDialog::Command* pCommand = reinterpret_cast<CHelperDialog::Command*>(m_CommandListBox.GetItemDataPtr( index ));

		if ( pCommand->m_Description.empty() )
		{
			m_CommandDescription.SetWindowTextA( "No description available." );
		}
		else
		{
			m_CommandDescription.SetWindowTextA( pCommand->m_Description.c_str() );
		}
	}
}

void CHelperDialog::ClearCommandListBox( void )
{
	while( m_CommandListBox.GetCount() )
	{
		m_CommandListBox.DeleteString(0);
	}
}

void CHelperDialog::ExecuteCommand( const std::string &commandName )
{
	// send the command name as a response
	s_SelectedCommand.SetValue( commandName );

	// Trigger event
	SetEvent( m_hEvents[0] );
}

void CHelperDialog::OnCbnSelchangeHelperFilter()
{
	CString filterName;
	int index = m_FilterComboBox.GetCurSel();
	if ( index != -1 )
	{
		// update to the selected filter
		m_FilterComboBox.GetLBText( index, filterName );
		SetActiveFilter( std::string(filterName) );
	}
	else
	{
		// clear the list box
		ClearCommandListBox();
	}
}

void CHelperDialog::OnCommandListDoubleClick(void)
{
	CString commandName;
	m_CommandListBox.GetText( m_CommandListBox.GetCurSel(), commandName );
	DisplayCommandDescription();
	ExecuteCommand( commandName.GetBuffer() );
}

void CHelperDialog::OnCommandSelectionChanged(void)
{
	DisplayCommandDescription();
}

void CHelperDialog::OnClose(void)
{
	// Clear the command list
	ClearCommandListBox();

	// Clear the filters
	m_Filters.empty();

	// Delete allocated commands
	std::list<CHelperDialog::Command*>::iterator itCommand = m_Commands.begin();
	while( itCommand != m_Commands.end() )
	{
		delete (*itCommand);
		itCommand++;
	}

	// Close dialog
	CWnd::OnClose();
}

void CHelperDialog::Filter::SetActiveFilter( CListBox &listBox )
{
	// clear the list box
	while ( listBox.GetCount() ) {
		listBox.DeleteString( 0 );
	}

	// Add all commands to the list box
	int index = -1;
	std::map<std::string, CHelperDialog::Command*>::iterator commandIt = m_Commands.begin();
	while ( commandIt != m_Commands.end() ) {
		index = listBox.AddString( commandIt->first.c_str() );
		if ( index < 0 )
		{
			throw std::exception( "Failed to change filter" );
		}
		else
		{
			listBox.SetItemDataPtr( index, commandIt->second );
		}

		commandIt++;
	}
}

void CHelperDialog::Filter::AddCommand( CHelperDialog::Command *pCommand )
{
	if ( m_Commands.find( pCommand->m_CommandName ) != m_Commands.end() )
	{
		// Can't map the same named command
		throw std::exception( std::string( pCommand->m_CommandName + std::string( ", should only be defined once." ) ).c_str() );
	}
	else
	{
		// Map the new command
		m_Commands[pCommand->m_CommandName] = pCommand;
	}
}
