/***
 *** ImplicitSend.cpp
 ***
 *** This is the source code for the standard implicit_send3.dll that is provided with the Profile Tuning Suite (PTS). It
 *** is provided to PTS users as a starting point for developing their own Implict Send DLL.
 ***
 *** Please see the "Automating - Using Implicit Send" document for more information. This document can be found in the
 *** same folder as this source code, or by selecting "Reference Documents" -> "Automating" from the PTS Start page.
 ***
 *** ----------
 ***
 *** This module provides the definitions for the five required functions that every Implicit Send DLL must provide. 
 ***
 *** ----------
 ***
 *** Copyright 2012, Bluetooth SIG, Inc.
 ***
 *** ----------
 ***
 *** Tom Allebrandi (tallebrandi@bluetooth.com)
 *** Bluetooth SIG, Inc.
 ***/


/****************************************************************************************************************************
 **
 ** Include Files
 **/

#define	WINVER	0x0501	// Windows XP
#include	<afxwin.h>
#include	<string>

#include	"implicitsendapi.h"
#include	"implicitsenddialog.h"
#include	"persistenttext.h"
#include	"helperdialog.h"

/****************************************************************************************************************************
 **
 ** Module Level Data
 **/
/*
 * See PersistenText.cpp for why we need this variable
 */
static PersistentText persistentText;


/****************************************************************************************************************************
 **
 ** InitImplicitSend()
 **
 ** Declaration: extern "C" bool WINAPI InitImplicitSend(void);
 **
 ** Parameters: None
 **
 ** Return values: true if successful
 **                false if not sucessful
 **
 ** This function is called during the initialization of an Executable Test Suite (ETS), just after the Implicit Send DLL 
 ** has been loaded into memory. It can be used to perform any initialization that might be needed before executing test
 ** cases.
 **
 ** If no initialization is required, the function can simply return a value of "true".
 **
 ** If the initializations failed, which would lead to the DLL not being usable, a "false" value should be returned. In this
 ** case, the ETS will be disabled.
 **/

extern "C" bool WINAPI InitImplicitSend()
{
/*
 * For the standard Implicit Send DLL we have nothing to initialize
 */
return(true);
}


/****************************************************************************************************************************
 **
 ** ImplicitStartTestCase()
 **
 ** Declaration: extern "C" void WINAPI ImplicitStartTestCase(std::string& strTestCaseName);
 **
 ** Parameters: A character string containing the name of the current test case
 **
 ** Return values: None
 **
 ** ImplicitStartTestCase() is called at the start of each test case execution. It provides the name of the test case that
 ** is starting. This function can be used to perform initializations that are needed at the start of every test case. The
 ** test case name allows the initialization process to be customized to specific test cases.
 **/

extern "C" void WINAPI ImplicitStartTestCase(
	std::string&	strTestCaseName)
{
/*
 * For the standard Implicit Send DLL we do not use the parameter
 */
UNREFERENCED_PARAMETER(strTestCaseName);
}


/****************************************************************************************************************************
 **
 ** ImplicitSendStyle()
 **
 ** Declaration: extern "C" char* WINAPI ImplicitSendStyle(std::string& strMmiText,UINT mmiStyle);
 **
 ** Parameters: strMmiText  Information about the MMI (Implicit Send) request being made
 **             mmiStyle  A value describing the type of request and the expected values
 **
 ** Return values: A pointer to a character string containing the information to be returned
 **                A NULL pointer if the request cannot be processed or no information is to be returned
 **
 ** This is the main routine for handling Implicit Send requests; the majority of interactions with the test operator or
 ** automated test environment will be handled by this function.
 **
 ** strMmiText will consist of two pieces
 **
 ** o) A message tag that uniquely identifies the message (see section 3.3 of "Automating - Using Implicit Send",
 **    Message tags).
 ** o) Message text that would normally be displayed to the test operator.
 **
 ** In the default Implicit Send DLL used by PTS (this DLL), the mmiStyle parameter is used to select how the information
 ** should be displayed to the user. Custom DLLs can use this information to determine how to process the strMmiText, the
 ** type of request that is being made, and the expected return values.
 **
 ** In most cases, the return value will be a pointer to a string containing the word OK. Some requests, such as those
 ** using MMI_Style_Edit1 expect a string of data  for example, a PIN code or a file name  as the return value.
 **
 ** The MMI_Style_Edit2 style provides a list of items in the strMmiText and expects one of those items to be returned.
 **
 ** For more information on the MMI styles see section 3.4 of "Automating - Using Implicit Send", MMI styles
 **/

extern "C" char* WINAPI ImplicitSendStyle(
	std::string&	strMmiText,
	UINT			mmiStyle)
{
/*
 * Validate the style value
 */
if ((mmiStyle != MMI_Style_Ok_Cancel1)
 && (mmiStyle != MMI_Style_Ok_Cancel2)
 && (mmiStyle != MMI_Style_Ok1)
 && (mmiStyle != MMI_Style_Yes_No1)
 && (mmiStyle != MMI_Style_Yes_No_Cancel1)
 && (mmiStyle != MMI_Style_Abort_Retry1)
 && (mmiStyle != MMI_Style_Edit1)
 && (mmiStyle != MMI_Style_Edit2)
 && (mmiStyle != MMI_Style_Helper))	// This style is used internally by PTS development and is intentionally undocumented
									// PTS users should not expect to see this style.
	{
	/*
	 * Display a message and pretend that the operator cancelled the dialog
	 */
	MessageBox(NULL,"Invalid style value passed to ImplicitSendStyle()!","Implicit Send",MB_ICONWARNING);
	persistentText.Store(NULL);
	goto ABORT;
	}

/*
 * In the standard Implicit Send DLL we interact with the user/test operator via a dialog box. The rest of this function
 * implements the use of the ImplicitSendDialog class found in ImplicitSendDialog.cpp
 *
 * If you are creating your own Implicit Send DLL remove the code from here to the comment that says "End Dialog Work" and
 * replace it with whatever you need to do.
 */

{// <---- Establishing a new variable scope to compartmentalize the dialog work
/*
 * Hand off to the internal development helper if that's what has been requested
 */
if (mmiStyle == MMI_Style_Helper)
	{
	// Helper MMI
	std::string result = RunHelperDialog( strMmiText, "" );
	if (result.empty())
		{
		persistentText.Store(NULL);
		}
	else
		{
		persistentText.Store(result .c_str());
		}

	return(const_cast<char*>(persistentText.GetPointer()));
	}

/*
 * MFC requires the following call in order to locate the dialog template in this DLL.
 */
AFX_MANAGE_STATE(AfxGetStaticModuleState());

/*
 * Declare an instance of the dialog
 */
ImplicitSendDialog	dialog;

/*
 * Send the parameters to the dialog box
 */
dialog.SetData(strMmiText,mmiStyle,"");

/*
 * Transfer control to the dialog
 */
switch (dialog.DoModal())
 {
 case ImplicitSendDialog::AffirmativeResponse:
	{
	/*
	 * This always means that the user has terminated the dialog in a positive manner. There may be user input or the users
	 * selection from a list depending on the style
	 */
	if ((mmiStyle == MMI_Style_Edit1)	// User input
	 || (mmiStyle == MMI_Style_Edit2))	// Selection from a list
		{
		/*
		 * Get the response data from the dialog
		 */
		persistentText.Store(dialog.GetData().c_str());
		}
	 else
		{
		/*
		 * Return an affirmative response.
		 */
		persistentText.Store("OK");
		}

	break;
	}

 case ImplicitSendDialog::NegativeResponse:
	{
	/*
	 * This always indicates some form of cancellation or termination request. There is no information to be returned
	 */
	persistentText.Store(NULL);
	break;
	}

 case ImplicitSendDialog::DialogOkTerminationRequestedByTestCase:
	{
	/*
	 * The current test case has requested that the dialog be terminated with an [Ok]. Return a response that might mean
	 * something to someone.
	 */
	persistentText.Store("Implicit Send Dialog Terminated");
	break;
	}

 case ImplicitSendDialog::DialogCancelTerminationRequestedByTestCase:
	{
	/*
	 * The current test case has requested that the dialog be terminated with a [Cancel]. We can't return a response since
	 * that would be treated as an [Ok].
	 */
	persistentText.Store(NULL);
	break;
	}

 }
}//<---- Terminating the variable scope used to compartmentalize the dialog work
/*
 * End Dialog Work
 */

ABORT:
/*
 * Return the pointer from the persistent storage class
 */
return(const_cast<char*>(persistentText.GetPointer()));
}


/****************************************************************************************************************************
 **
 ** ImplicitSendPinCode()
 **
 ** Declaration: extern "C" char* WINAPI ImplicitSendPinCode(void);
 **
 ** Parameters: None
 **
 ** Return values: A pointer to a character string containing a PIN code to be returned
 **                A NULL pointer if the request cannot be processed or no PIN code is to be returned
 **
 ** This is a special case function that is only used when a dynamic PIN code is needed.
 **
 ** It is not currently used in PTS, but it might be used in the future.
 **/

extern "C" char* WINAPI ImplicitSendPinCode(void)
{
/*
 * As mentioned in section 3.2.5 of "Automating - Using Implicit Send", this function can be implemented as follows
 */
std::string	strPrompt = "Please enter a PIN Code:";
return(ImplicitSendStyle(strPrompt,MMI_Style_Edit1));
}


/****************************************************************************************************************************
 **
 ** ImplicitTestCaseFinished()
 **
 ** Declaration: extern "C" void WINAPI ImplicitTestCaseFinished(void);
 **
 ** Parameters: None
 **
 ** Return values: None
 **
 ** ImplicitTestCaseFinished() is called at the end of test case execution. It may be used to perform any cleanup that is
 ** needed, or to undo operations that were performed during ImplicitStartTestCase().
 **/

extern "C" void WINAPI ImplicitTestCaseFinished()
{
/*
 * For the standard Implicit Send DLL we have nothing to do here
 */
}

/****************************************************************************************************************************
 **
 ** ImplicitSendStyleEx()
 **
 ** Declaration: extern "C" char* WINAPI ImplicitSendStyle(std::string& strMmiText,UINT mmiStyle, std::string& strBdAddr);
 **
 ** Parameters: strMmiText  Information about the MMI (Implicit Send) request being made
 **             mmiStyle  A value describing the type of request and the expected values
 **
 ** Return values: A pointer to a character string containing the information to be returned
 **                A NULL pointer if the request cannot be processed or no information is to be returned
 **
 ** This is the main routine for handling Implicit Send requests; the majority of interactions with the test operator or
 ** automated test environment will be handled by this function.
 **
 ** strMmiText will consist of two pieces
 **
 ** o) A message tag that uniquely identifies the message (see section 3.3 of "Automating - Using Implicit Send",
 **    Message tags).
 ** o) Message text that would normally be displayed to the test operator.
 **
 ** In the default Implicit Send DLL used by PTS (this DLL), the mmiStyle parameter is used to select how the information
 ** should be displayed to the user. Custom DLLs can use this information to determine how to process the strMmiText, the
 ** type of request that is being made, and the expected return values.
 **
 ** In most cases, the return value will be a pointer to a string containing the word OK. Some requests, such as those
 ** using MMI_Style_Edit1 expect a string of data  for example, a PIN code or a file name  as the return value.
 **
 ** The MMI_Style_Edit2 style provides a list of items in the strMmiText and expects one of those items to be returned.
 **
 ** For more information on the MMI styles see section 3.4 of "Automating - Using Implicit Send", MMI styles
 **/

extern "C" char* WINAPI ImplicitSendStyleEx(std::string& strMmiText, UINT	mmiStyle, std::string& strBdAddr)
{
	/*
	 * Validate the style value
	 */
	if ((mmiStyle != MMI_Style_Ok_Cancel1) && (mmiStyle != MMI_Style_Ok_Cancel2) && (mmiStyle != MMI_Style_Ok1)
		&& (mmiStyle != MMI_Style_Yes_No1) && (mmiStyle != MMI_Style_Yes_No_Cancel1) && (mmiStyle != MMI_Style_Abort_Retry1)
		&& (mmiStyle != MMI_Style_Edit1) && (mmiStyle != MMI_Style_Edit2) && (mmiStyle != MMI_Style_Helper))	// This style is used internally by PTS development and is intentionally undocumented PTS users should not expect to see this style.
	{
		/*
		 * Display a message and pretend that the operator cancelled the dialog
		 */
		MessageBox(NULL,"Invalid style value passed to ImplicitSendStyle()!","Implicit Send",MB_ICONWARNING);
		persistentText.Store(NULL);
		goto ABORT;
	}

	/*
	 * In the standard Implicit Send DLL we interact with the user/test operator via a dialog box. The rest of this function
	 * implements the use of the ImplicitSendDialog class found in ImplicitSendDialog.cpp
	 *
	 * If you are creating your own Implicit Send DLL remove the code from here to the comment that says "End Dialog Work" and
	 * replace it with whatever you need to do.
	 */

	{// <---- Establishing a new variable scope to compartmentalize the dialog work
		/*
		 * Hand off to the internal development helper if that's what has been requested
		 */
		if (mmiStyle == MMI_Style_Helper)
		{
			// Helper MMI
			std::string result = RunHelperDialog(strMmiText, strBdAddr);
			if (result.empty())
			{
				persistentText.Store(NULL);
			}
			else
			{
				persistentText.Store(result .c_str());
			}

			return(const_cast<char*>(persistentText.GetPointer()));
		}

		/*
		 * MFC requires the following call in order to locate the dialog template in this DLL.
		 */
		AFX_MANAGE_STATE(AfxGetStaticModuleState());

		/*
		 * Declare an instance of the dialog
		 */
		ImplicitSendDialog	dialog;

		/*
		 * Send the parameters to the dialog box
		 */
		dialog.SetData(strMmiText,mmiStyle,strBdAddr);

		/*
		 * Transfer control to the dialog
		 */
		switch (dialog.DoModal())
		{
			case ImplicitSendDialog::AffirmativeResponse:
			{
				/*
				 * This always means that the user has terminated the dialog in a positive manner. There may be user input or the users
				 * selection from a list depending on the style
				 */
				if ((mmiStyle == MMI_Style_Edit1)	// User input
				 || (mmiStyle == MMI_Style_Edit2))	// Selection from a list
				{
					/*
					 * Get the response data from the dialog
					 */
					persistentText.Store(dialog.GetData().c_str());
				}
				else
				{
					/*
					 * Return an affirmative response.
					 */
					persistentText.Store("OK");
				}
				break;
			}

			case ImplicitSendDialog::NegativeResponse:
			{
				/*
				 * This always indicates some form of cancellation or termination request. There is no information to be returned
				 */
				persistentText.Store(NULL);
				break;
			}

			case ImplicitSendDialog::DialogOkTerminationRequestedByTestCase:
			{
				/*
				 * The current test case has requested that the dialog be terminated with an [Ok]. Return a response that might mean
				 * something to someone.
				 */
				persistentText.Store("Implicit Send Dialog Terminated");
				break;
			}

			case ImplicitSendDialog::DialogCancelTerminationRequestedByTestCase:
			{
				/*
				 * The current test case has requested that the dialog be terminated with a [Cancel]. We can't return a response since
				 * that would be treated as an [Ok].
				 */
				persistentText.Store(NULL);
				break;
			}

		}
	}//<---- Terminating the variable scope used to compartmentalize the dialog work
	/*
	 * End Dialog Work
	 */

	ABORT:
	/*
	 * Return the pointer from the persistent storage class
	 */
	return(const_cast<char*>(persistentText.GetPointer()));
}


/****************************************************************************************************************************
 **
 ** ImplicitSendPinCodeEx()
 **
 ** Declaration: extern "C" char* WINAPI ImplicitSendPinCode(void);
 **
 ** Parameters: None
 **
 ** Return values: A pointer to a character string containing a PIN code to be returned
 **                A NULL pointer if the request cannot be processed or no PIN code is to be returned
 **
 ** This is a special case function that is only used when a dynamic PIN code is needed.
 **
 ** It is not currently used in PTS, but it might be used in the future.
 **/

extern "C" char* WINAPI ImplicitSendPinCodeEx(std::string& strBdAddr)
{
/*
 * As mentioned in section 3.2.5 of "Automating - Using Implicit Send", this function can be implemented as follows
 */
std::string	strPrompt = "Please enter a PIN Code:";
return(ImplicitSendStyleEx(strPrompt,MMI_Style_Edit1,strBdAddr));
}

