/***
 *** MyCStatic.Cpp
 ***
 *** An extended CStatic control that provides a way to compute an optimal size based on its current text.
 ***
 *** ----------
 ***
 *** Originally developed by me at Frontline Test Equipment, Inc.
 ***
 *** ----------
 ***
 *** Tom Allebrandi (tallebrandi@bluetooth.com/wyrles@ytram.com/tom@ytram.com)
 ***/

/****************************************************************************************************************************
 **
 ** Includes
 **/

#define	WINVER	0x0501
#include	<afxwin.h>
#include	<math.h>

#include	"mycstatic.h"

#ifdef	MYCSTATIC_TEST_PROGRAM
#include	<stdio.h>
#endif

/****************************************************************************************************************************
 **
 ** Constants
 **/

enum	{
		TEXT_FORMAT = (DT_LEFT | DT_WORDBREAK | DT_EXPANDTABS)
		};

/****************************************************************************************************************************
 **
 ** MyCEdit Message Map
 **/

BEGIN_MESSAGE_MAP(MyCEdit, CEdit)
	ON_WM_PAINT()
END_MESSAGE_MAP()


/****************************************************************************************************************************
 **
 ** MyCStatic::ComputeGoodSize()
 **/

void MyCEdit::ComputeGoodSize(
	int&	newWidth,
	int&	newHeight,
	int		xScale,
	int		yScale,
	bool	computeDeltaSize)
{
	double	area;
	double	computedHeight;
	double	computedWidth;
	CString	csText;
	double	currentHeight;
	double	currentWidth;
	double	m;
	double	oneLineHeight;
	double	oneLineWidth;
	RECT	rect;

#ifdef	MYCSTATIC_TEST_PROGRAM
printf_s("xScale = %d, yScale = %d, computeDeltaSize = %s\n",xScale,yScale,computeDeltaSize?"true":"false");
#endif

/*
 * How high would one line of text be?
 */
GetWindowText(csText);
SetWindowText("M");
GetClientRect(&rect);
myDrawText(TEXT_FORMAT|DT_CALCRECT,&rect);
oneLineWidth = rect.right;
oneLineHeight = rect.bottom;
SetWindowText(csText);

#ifdef	MYCSTATIC_TEST_PROGRAM
printf_s("oneLineWidth = %f, oneLineHeight = %f\n",oneLineWidth,oneLineHeight);
#endif

/*
 * Set the output size to 0
 */
newWidth = 0;
newHeight = 0;

/*
 * What is the current size of the control?
 */
GetClientRect(&rect);
currentWidth = rect.right;
currentHeight = rect.bottom;

#ifdef	MYCSTATIC_TEST_PROGRAM
printf_s("currentWidth = %f,currentHeight = %f\n",currentWidth,currentHeight);
#endif

/*
 * Compute a new size for the control
 */
myDrawText(TEXT_FORMAT|DT_CALCRECT,&rect);
computedWidth = rect.right;
computedHeight = rect.bottom;

#ifdef	MYCSTATIC_TEST_PROGRAM
printf_s("computedWidth = %f, computedHeight = %f\n",computedWidth,computedHeight);
#endif

/*
 * What are we supposed to do?
 */
if ((xScale <= 1)
 && (yScale <= 1))
	{
	/*
	 * No scaling needed
	 */
	#ifdef	MYCSTATIC_TEST_PROGRAM
	printf_s("No scaling requested\n");
	#endif
	}

 else if (xScale <= 1)
	{
	/*
	 * Leave the width alone and compute a new height
	 */
	area = computedWidth * computedHeight;
	computedWidth = currentWidth;
	computedHeight = area / currentWidth;

	#ifdef	MYCSTATIC_TEST_PROGRAM
	printf_s("Keep the current width, compute new height: area = %f, computedWidth = %f, computedHeight = %f\n",area,
		computedWidth,computedHeight);
	#endif
	}

 else if (yScale <= 1)
	{
	/*
	 * Leave the height alone and compute a new width
	 */
	area = computedWidth * computedHeight;
	computedWidth = area / currentHeight;
	computedHeight = currentHeight;

	#ifdef	MYCSTATIC_TEST_PROGRAM
	printf_s("Keep the current height, compute new width: area = %f, computedWidth = %f, computedHeight = %f\n",area,
		computedWidth,computedHeight);
	#endif
	}

 else
	{
	/*
	 * Rescale the dimensions of the window based on the scaling values
	 *
	 * width * height = <window area> = (xScale * m) * (yScale * m)
	 *
	 * solve for m
	 *
	 */
	m = sqrt((computedWidth * computedHeight) / ((double) xScale * (double) yScale));

	/*
	 * From m, compute the new size
	 */
	computedWidth = (double) xScale * m;
	computedHeight = (double) yScale * m;

	#ifdef	MYCSTATIC_TEST_PROGRAM
	area = computedWidth * computedHeight;
	printf_s("Scale the window size: m = %f, area = %f, computedWidth = %f, computedHeight = %f\n",m,area,computedWidth,
		computedHeight);
	#endif

	/*
	 * Adjust to fit
	 */
	::SetRect(&rect,0,0,(int) ceil(computedWidth),(int) ceil(computedHeight));
	myDrawText(TEXT_FORMAT|DT_CALCRECT,&rect);
	computedWidth = rect.right;
	computedHeight = rect.bottom;

	#ifdef	MYCSTATIC_TEST_PROGRAM
	area = computedWidth * computedHeight;
	printf_s("Adjust to fit: area = %f, computedWidth = %f, computedHeight = %f\n",area,computedWidth,computedHeight);
	#endif
	}

/*
 * Return the absolute new size or the delta size?
 */
if (computeDeltaSize)
	{
	/*
	 * Delta. Adjust the new size by subtracting the current size
	 */
	computedWidth -= currentWidth;
	computedHeight -= currentHeight;

	#ifdef	MYCSTATIC_TEST_PROGRAM
	printf_s("Compute delta size: computedWidth = %f, computedHeight = %f\n",computedWidth,computedHeight);
	#endif
	}

/*
 * Set the output size to the computed new size. For some reason we need to add about one and a half lines worth of height
 * to the computed value
 */
newWidth = (int) ceil(computedWidth);
newHeight = (int) ceil(computedHeight + (1.5 * oneLineHeight));

#ifdef	MYCSTATIC_TEST_PROGRAM
printf_s("newWidth = %d, newHeight = %d\n",newWidth,newHeight);
#endif
}

/*******************************************************************************
 **
 ** MyCEdit::myDrawText()
 **/

void MyCEdit::myDrawText(
	DWORD		format,
	RECT*		pRect)
{
	CString		csText;
	CDC			*pDc;
	CRect		r;

/*
 * Use the supplied rectangle or the current client rectangle?
 */
if (pRect != NULL)
	{
	/*
	 * Caller supplied
	 */
	r = *pRect;
	}
 else
	{
	/*
	 * Client window
	 */
	GetClientRect(r);
	}

/*
 * Just calculating the size
 */
if ((format & DT_CALCRECT) != 0)
	{
	/*
	 * We are computing whether or not the text will fit
	 */
	pDc = GetDC();

	/*
	 * Get the current text in the control
	 */
	GetWindowText(csText);

	/*
	 * The dialog box font is not currently active in the device context
	 */
	pDc->SelectObject(GetOwner()->GetFont());

	/*
	 * "Draw"
	 */
	pDc->SetBkMode(TRANSPARENT);
	pDc->DrawText(csText, &r, format);
	}

/*
 * Actually paint, call standard routine because CEdit is quite complicated
 */
else
	{
	CEdit::OnPaint();
	}

/*
 * Return the rectangle if one was supplied
 */
if (pRect != NULL)
	{
	*pRect = r;
	}
}


/*******************************************************************************
 **
 ** MyCStatic::OnPaint()
 **/

void MyCEdit::OnPaint()
{
/*
 * Put the current text onto the display
 */
myDrawText(TEXT_FORMAT,NULL);
}

/****************************************************************************************************************************
 ****************************************************************************************************************************
 ***************************************************************************************************************************/

/****************************************************************************************************************************
 **
 ** Test program
 **/

#ifdef	MYCSTATIC_TEST_PROGRAM

/****************************************************************************************************************************
 *
 * MFC requires an instance of a CWinApp object to hold state information in.
 */
CWinApp thisDll;

/****************************************************************************************************************************
 **
 ** We need a dialog box for testing
 **/

#include	"resource.h"

class Dialog : public CDialog
{
/*
 * Construction/Destruction
 */
public:
			Dialog(void) : CDialog(IDD_DIALOG)
				{
				};
			~Dialog(void)
				{
				};

/*
 * Interface
 */
public:
	void	SetData(int xScale,int yScale,const char* pszText)
				{
				m_xScale = xScale;
				m_yScale = yScale;
				m_pszText = pszText;
				};

/*
 * Class internal functions
 */
private:
	/*
	 * Window message handlers
	 */
	void	DoDataExchange(CDataExchange* pDx)
				{
				CDialog::DoDataExchange(pDx);
				DDX_Control(pDx,IDC_BEFORE,m_cBefore);
				DDX_Control(pDx,IDC_AFTER,m_cAfter);
				};

	void	OnCancel(void)
				{
				EndDialog(IDCANCEL);
				};
	
	BOOL	OnInitDialog(void)
				{
				CDialog::OnInitDialog();

				SetWindowText("MyCStatic Test");
				m_cBefore.SetWindowText(m_pszText);
				m_cAfter.SetWindowText(m_pszText);

				/*
				 * Using MyCStatic...
				 */
				// Compute the best size for the control as a delta from its current size
				int newWidth;
				int newHeight;
				m_cAfter.ComputeGoodSize(newWidth,newHeight,m_xScale,m_yScale,false);

				// Get the current size and position of the control
				RECT rBefore;
				RECT rAfter;
				m_cAfter.GetWindowRect(&rBefore);

				// The left and top coordinates for MoveWindow() are relative to the client area of the dialog not the upper
				// left hand corner the GetWindowRect() gave us. Correct for that.
				rAfter.left = rBefore.left - GetSystemMetrics(SM_CXDLGFRAME);									// The width of the window frame
				rAfter.top  = rBefore.top - (GetSystemMetrics(SM_CYDLGFRAME) + GetSystemMetrics(SM_CYCAPTION));	// The height of the window frame and title bar

				// Adjust the size
				rAfter.right = rAfter.left + newWidth;
				rAfter.bottom = rAfter.top + newHeight;

				// Resize the window to its new dimensions
				m_cAfter.MoveWindow(rAfter.left,rAfter.top,newWidth,newHeight,false);

				// We may also need to resize the dialog
				int deltaWidth = (rAfter.right - rAfter.left) - (rBefore.right - rBefore.left);
				int deltaHeight = (rAfter.bottom - rAfter.top) - (rBefore.bottom - rBefore.top);
				if ((deltaWidth > 0)
				 || (deltaHeight > 0))
					{
					RECT r;
					GetWindowRect(&r);
					r.right += deltaWidth;
					r.bottom += deltaHeight;
					SetWindowPos(&wndTop,0,0,r.right-r.left,r.bottom-r.top,SWP_NOMOVE|SWP_NOACTIVATE);
					}

				return(FALSE);
				};


	void	OnOK(void)
				{
				EndDialog(IDOK);
				};

	DECLARE_MESSAGE_MAP();

/*
 * Class data
 */
private:
	/*
	 * Controls
	 */
	CStatic		m_cBefore;
	MyCStatic	m_cAfter;

	/*
	 * Other data
	 */
	int			m_xScale;
	int			m_yScale;
	const char*	m_pszText;
};

BEGIN_MESSAGE_MAP(Dialog,CDialog)
END_MESSAGE_MAP()

/****************************************************************************************************************************
 **
 ** Main Program
 **/

#include	<string>

C_ASSERT(sizeof(int) == sizeof(DWORD));

DWORD main(
	int						argc,
	char**					argv)
{
	Dialog					dlg;
	FILE*					f = NULL;
	std::string::size_type	index;
	char*					pszEnd;
	char*					pszXScale;
	char*					pszYScale;
	std::string				strText;
	char					szBuffer[1024];
	int						xScale;
	int						yScale;

/*
 * Check for the proper number of parameters
 */
if ((argc != 3)
 && (argc != 4))
	{
	fprintf_s(stderr,"Usage: %s <xScale> <yScale> [<text>]\n",argv[0]);
	goto ABORT;
	}

/*
 * For convenience, map the arguments to useful names
 */
pszXScale = argv[1];
pszYScale = argv[2];
if (argc == 4)
	{
	strText = std::string(argv[3]);
	}

/*
 * Make sure that the first two parameters are integers
 */
xScale = strtol(pszXScale,&pszEnd,10);
if (*pszEnd != '\0')
	{
	fprintf_s(stderr,"xScale value is not an integer\n");
	fprintf_s(stderr,"\n");
	fprintf_s(stderr,"Usage: %s <xScale> <yScale> <text>\n",argv[0]);
	goto ABORT;
	}

yScale = strtol(pszYScale,&pszEnd,10);
if (*pszEnd != '\0')
	{
	fprintf_s(stderr,"yScale value is not an integer\n");
	fprintf_s(stderr,"\n");
	fprintf_s(stderr,"Usage: %s <xScale> <yScale> <text>\n",argv[0]);
	goto ABORT;
	}

/*
 * Make sure that we have some text
 */
if (strText.empty())
	{
	strText =
		"This is a test of the Emergency Broadcast System. The broadcasters of your area in voluntary cooperation with the"
		" Federal, State and local authorities have developed this system to keep you informed in the event of an emergency."
		" If this had been an actual emergency, the Attention Signal you just heard would have been followed by official"
		" information, news or instructions. This station serves the Seattle, Washington area.\\n"
		"\\n"
		"This concludes this test of the Emergency Broadcast System.";
	}

/*
 * Allow the text to come from a file
 */
if (strText[0] == '@')
	{
	/*
	 * Open the file
	 */
	if (fopen_s(&f,strText.substr(1).c_str(),"r") != 0)
		{
		fprintf_s(stderr,"Failed to open %s\n",strText.substr(1).c_str());
		goto ABORT;
		}

	/*
	 * Slurp the file into the string variable
	 */
	strText.clear();
	while (fgets(szBuffer,_countof(szBuffer),f) != NULL)
		{
		strText += std::string(szBuffer);
		}

	/*
	 * Close the file
	 */
	fclose(f);
	f = NULL;
	}

/*
 * Allow newlines in the text indicated by a "\n" pair
 */
while ((index=strText.find("\\n")) != std::string::npos)
	{
	strText.replace(index,2,"\n");
	}

/*
 * Initialize MFC
 */
if (!AfxWinInit(GetModuleHandle(NULL),NULL,GetCommandLine(),0))
	{
	fprintf_s(stderr,"Failed to initialize MFC\n");
	goto ABORT;
	}

/*
 * Show the dialog. It has two text boxes: one before scaling and one after scaling
 */
dlg.SetData(xScale,yScale,strText.c_str());
dlg.DoModal();

ABORT:
if (f != NULL)
	{
	fclose(f);
	}

return(ERROR_SUCCESS);
}
#endif
