/***
 *** PersistentText.cpp
 ***
 *** This is the source code for the standard implicit_send3.dll that is provided with the Profile Tuning Suite (PTS). It
 *** is provided to PTS users as a starting point for developing their own Implict Send DLL.
 ***
 *** Please see the "Automating - Using Implicit Send" document for more information. This document can be found in the
 *** same folder as this source code, or by selecting "Reference Documents" -> "Automating" from the PTS Start page.
 ***
 *** ----------
 ***
 *** As discussed in Section 3.2.4.1 of "Automating - Using Implicit Send", the return value from ImplicitSendStyle() and
 *** ImplicitSendPinCode() needs to persist after we exit those functions. If your implementation has a way to return
 *** pointers to literal strings or static text, then you don't need to be concered about this issue.
 ***
 *** However, if your implementation does something like the dialog work in the standard implementation of
 *** ImplicitSendStyle() (and ImplicitSendPinCode()) the text to be returned ends up in a local variable that will be
 *** destroyed upon exit from the routine. So, you need to find a way to make the data live on.
 ***
 *** The implementation we have used here is to make a copy of the text and store a pointer to it in a module level static
 *** variable. That way we can safely return a pointer to text that lives on after the function has completed.
 ***
 *** For convenience we have a used a C++ class as a container for the pointer then created a module level instance of the
 *** class. That gives us a clean way to make sure that the pointer is initialized properly when the DLL is loaded and that
 *** any space held by the pointer is cleaned up when DLL is unloaded.
 ***
 ***
 *** This file defines the implementation of the class.
 ***
 *** ----------
 ***
 *** Copyright 2012, Bluetooth SIG, Inc.
 ***
 *** ----------
 ***
 *** Tom Allebrandi (tallebrandi@bluetooth.com)
 *** Bluetooth SIG, Inc.
 ***/

/****************************************************************************************************************************
 **
 ** Includes
 **/
#include	<malloc.h>
#include	<string.h>

#include	"persistenttext.h"


/****************************************************************************************************************************
 **
 ** PersistentText::PersistentText()
 **
 ** Class constructor
 **/

PersistentText::PersistentText()
{
/*
 * Initialize the pointer
 */
m_pszText = NULL;
}


/****************************************************************************************************************************
 **
 ** PersistentText::~PersistentText()
 **
 ** Class destructor
 **/

PersistentText::~PersistentText()
{
/*
 * Make sure that any dynamic memory is released
 */
if (m_pszText != NULL)
	{
	free(const_cast<char*>(m_pszText));
	m_pszText = NULL;
	}
}

/****************************************************************************************************************************
 **
 ** PersistentText::GetPointer()
 **
 ** Retreive the pointer to the stored text
 **/

const char* PersistentText::GetPointer(void)
{
return(m_pszText);
}


/****************************************************************************************************************************
 **
 ** PersistentText::Store()
 **
 ** Store some text, replacing what may already be stored
 **/

void PersistentText::Store(
	const char*	pszText)
{
/*
 * Do we have anything currently stored?
 */
if (m_pszText != NULL)
	{
	/*
	 * Yes, destroy it
	 */
	free(const_cast<char*>(m_pszText));
	m_pszText = NULL;
	}

/*
 * Do we have anything to store?
 */
if (pszText != NULL)
	{
	/*
	 * Yes, make a copy and store the pointer to it
	 */
	m_pszText = _strdup(pszText);
	// We really should do something here if this _strdup() fails...
	}
}
