/**************************************************************************
***************************************************************************
**
** File:					PTSControlClient.cpp
**
** Description:				Sample application to show using the PTSControl API
**
** Owner:					Bluetooth SIG, Inc.
**
** Created:					2011-01-03
**
** Contributor:				Microsoft Corporation
**
** Notes:					For further details of PTSControl, please refer to document "Extended_Automating.pdf"
**
****************************************************************************
*
*   2011 Bluetooth SIG, Inc. All rights reserved
* 
****************************************************************************/

#include "stdafx.h"
#include <vector>
#include "PTSControl.h"
#import <msxml6.dll> exclude("IStream", "ISequentialStream", "_LARGE_INTEGER", "_ULARGE_INTEGER", "tagSTATSTG", "_FILETIME")
#include <strsafe.h>	// Must be last or you get a pile of "C4995: 'function': name was marked as #pragma deprecated" warnings
#include <string>
#include <Psapi.h>
typedef struct
{
	DWORD	processId;
	HWND	hWnd;
}	getWnd_parameters;

static bool ClosePTS(CComPtr<IPTSControl> &pPTS);
static HWND GetHwndFromProccess(HANDLE hProcess);
static BOOL CALLBACK GetProcessWndEnum(HWND hwnd, LPARAM lParam);

/* ===========================================================================
 * Function: 	doLog
 * Description: Display log output on the console window. Since this is a module level function that is not part of the CPTSControlClientLogger class, it can be used at any time
 * Returns:		HRESULT - result code
 * Arguments:	IN 	int - log type (See PTSControl.h for all the types. Note that PTSControl.h does not belong to this project but it is located within the same file directory of this file.)
 *				IN  LPCWSTR - log type
 *				IN  LPCWSTR - time
 *				IN	LPCWSTR - message to be logged
 * =========================================================================== */

HANDLE hStdout = INVALID_HANDLE_VALUE;
WORD wOldColorAttrs = 0;
ULONGLONG ptsBdAddr = 0;

#define	LOCAL_MESSAGE	100
#define	LOCAL_ERROR		101

/* ===========================================================================
* Function: 	VariantToStringVector
* Description: Helper function to convert a VARIANT holding array of BSTRs to a vector of string-type objects
* Returns:		S_OK in case of succes, error code in case of failure
* Arguments:	IN	vt		- VARIANT holding array of BSTRs
*				OUT	vResult	- vector of string-type objects
* =========================================================================== */
template <class T>
HRESULT VariantToStringVector(const VARIANT& vt, std::vector<T>& vResult)
{
	if (!(vt.vt && VT_BSTR) || !(vt.vt & VT_ARRAY))
	{
		return DISP_E_TYPEMISMATCH;
	}

	SAFEARRAY* psArray;
	if (vt.vt & VT_BYREF)
	{
		psArray = *(vt.pparray);
	}
	else
	{
		psArray = vt.parray;
	}

	long lLBound, lUBound;
	SafeArrayGetLBound(psArray, 1, &lLBound);
	SafeArrayGetUBound(psArray, 1, &lUBound);
	for (long i = lLBound; i <= lUBound; ++i)
	{
		CComBSTR bstr;
		HRESULT hrGetElem = SafeArrayGetElement(psArray, &i, &bstr);
		if (SUCCEEDED(hrGetElem))
		{
			T item = bstr;
			vResult.push_back(item);
		}
	}

	return S_OK;
}

static HRESULT doLog(int logType, LPCWSTR szLogType, LPCWSTR szTime, LPCWSTR pszMessage)
{
    BOOL bRes = TRUE;
    switch(logType)
    {
    case PTS_LOGTYPE_IMPLICIT_SEND:
        bRes = SetConsoleTextAttribute(hStdout, FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_INTENSITY);
        break;
	case PTS_LOGTYPE_INFRASTRUCTURE:
    case PTS_LOGTYPE_ERROR:
        bRes = SetConsoleTextAttribute(hStdout, FOREGROUND_RED | FOREGROUND_INTENSITY);
        break;
    case PTS_LOGTYPE_PRELIMINARY_VERDICT:
    case PTS_LOGTYPE_FINAL_VERDICT:
        bRes = SetConsoleTextAttribute(hStdout, FOREGROUND_GREEN | FOREGROUND_INTENSITY);
        break;
    case PTS_LOGTYPE_START_TEST:
    case PTS_LOGTYPE_END_TEST:
        bRes = SetConsoleTextAttribute(hStdout, FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY);
        break;
    case PTS_LOGTYPE_MESSAGE:
    case PTS_LOGTYPE_SEND_EVENT:
    case PTS_LOGTYPE_RECEIVE_EVENT:
        bRes = SetConsoleTextAttribute(hStdout, FOREGROUND_GREEN | FOREGROUND_BLUE);
        break;
	case PTS_LOGTYPE_EVENT_SUMMARY:
        bRes = SetConsoleTextAttribute(hStdout, FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY);
        break;
	case LOCAL_MESSAGE:
        bRes = SetConsoleTextAttribute(hStdout, FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE);
		break;
	case LOCAL_ERROR:
        bRes = SetConsoleTextAttribute(hStdout, FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_INTENSITY);
		break;
    default:
        bRes = SetConsoleTextAttribute(hStdout, wOldColorAttrs);
        break;
    }

    if (!bRes)
    {
        return HRESULT_FROM_WIN32(GetLastError());
    }

    DWORD dwWritten;
    if (!WriteConsole(hStdout, pszMessage, lstrlen(pszMessage), &dwWritten, NULL))
    {
        return HRESULT_FROM_WIN32(GetLastError());
    }

	if (pszMessage[lstrlen(pszMessage)-1] != L'\n')
	{
	    if (!WriteConsole(hStdout, L"\r\n", lstrlen(L"\r\n"), &dwWritten, NULL))
		{
	        return HRESULT_FROM_WIN32(GetLastError());
	    }
	}

    return S_OK;
}

class CPTSControlClientLogger 
    : public IPTSControlClientLogger 
    , public IPTSImplicitSendCallback
	, public IPTSImplicitSendCallbackEx
{
public:
    CPTSControlClientLogger(HRESULT &hr);
    
    // IPTSControlClientLogger
    HRESULT STDMETHODCALLTYPE Log(PTS_LOGTYPE logType, LPCWSTR szLogType, LPCWSTR szTime, LPCWSTR pszMessage) override;
    
    // IPTSImplicitSendCallback
	HRESULT STDMETHODCALLTYPE OnSend(LPCWSTR pszProjectName, WORD wID, LPCWSTR pszTestCase, LPCWSTR pszDescriptor) override;

	// IPTSImplicitSendCallbackEx
	HRESULT STDMETHODCALLTYPE OnImplicitSend(LPCWSTR pszProjectName,WORD wID,LPCWSTR pszTestCase,LPCWSTR pszDescription,DWORD style,LPWSTR pszResponse,DWORD responseSize,BOOL* pbResponseIsPresent);

    //  IUnknown
    ULONG STDMETHODCALLTYPE AddRef();
    ULONG STDMETHODCALLTYPE Release();
    HRESULT STDMETHODCALLTYPE QueryInterface(REFIID iid, void **ppv);
private:
	~CPTSControlClientLogger();
    LONG m_cRef;
};

CPTSControlClientLogger::CPTSControlClientLogger(HRESULT &hr)
: m_cRef(0)
{
    if (FAILED(hr)) { return; }
}

CPTSControlClientLogger::~CPTSControlClientLogger()
{
}


/* ===========================================================================
 * Function: 	CPTSControlClientLogger::Log
 * Description: Default logging function capturing all logs generated by PTS while running test cases.
 * Returns:		HRESULT - result code
 * Arguments:	IN 	PTS_LOGTYPE - log type (See PTSControl.h for all the types. Note that PTSControl.h does not belong to this project but it is located within the same file directory of this file.)
 *				IN  LPCWSTR - log type
 *				IN  LPCWSTR - time
 *				IN	LPCWSTR - message to be logged
 * =========================================================================== */
HRESULT CPTSControlClientLogger::Log(PTS_LOGTYPE logType, LPCWSTR szLogType, LPCWSTR szTime, LPCWSTR pszMessage)
{
	// Hand off to the module's doLog() function defined above
	return(doLog(logType,szLogType,szTime,pszMessage));
}

/* ===========================================================================
 * Function: 	CPTSControlClientLogger::OnSend
 * Description: Default callback function from CPTSControlClientLogger replacing CImplicit_sendApp::ImplicitSend.
 *				This function should be used when RegisterImplicitSendCallback is called.
 * Returns:		HRESULT - result code
 * Arguments:	IN 	LPCWSTR - name of project / test suite
 *				IN  WORD - ID of MMI
 *				IN  LPCWSTR - name of test case
 *				IN	LPCWSTR - description of the MMI
 * =========================================================================== */
HRESULT CPTSControlClientLogger::OnSend(LPCWSTR pszProjectName, WORD wID, LPCWSTR pszTestCase, LPCWSTR pszDescription)
{
	// Replace the following message box and customize it for your device
	wchar_t message[2048];
	swprintf_s(message,_countof(message),
		L"CPTSControlClientLogger::OnSend() has been called with the following parameters:\r\n"
		L"\r\n"
		L"pszProjectName = \"%s\"\r\n"
		L"wID = %d\r\n"
		L"pszTestCase = \"%s\"\r\n"
		L"pszDescription = \"%s\"",pszProjectName,wID,pszTestCase,pszDescription);
	MessageBox(NULL,message,L"CPTSControlClientLogger::OnSend()",MB_OK);


	// Replace the following function call to Log if you want to change the way it is logged.
	HRESULT hr = Log(PTS_LOGTYPE_IMPLICIT_SEND, L"", L"", L"Implicit Send callback OnSend() has been called.\n");
	return hr;
}

/* ===========================================================================
 * Function: 	CPTSControlClientLogger::OnImplicitSend
 * Description: Default callback function from CPTSControlClientLogger replacing CImplicit_sendApp::ImplicitSend.
 *				This function should be used when RegisterImplicitSendCallbackEx is called.
 * Returns:		HRESULT - result code
 * Arguments:	IN 	LPCWSTR - name of project / test suite
 *				IN  WORD - ID of MMI
 *				IN  LPCWSTR - name of test case
 *				IN	LPCWSTR - description of the MMI
 *				IN  DWORD - style of the MMI
 *              IN,OUT LPWSTR - response text
 *              IN  DWORD - length of response text
 * =========================================================================== */

// This list is from ImplicitSendAPI.h, normally found in the "Bluetooth PTS\Custom\Implicit_Send" folder
enum ImplicitSendMessageStyles
		{
		MMI_Style_Ok_Cancel1 =     0x11041,	// Simple prompt           | OK, Cancel buttons      | Default: OK
		MMI_Style_Ok_Cancel2 =     0x11141,	// Simple prompt           | Cancel button           | Default: Cancel
		MMI_Style_Ok1 =            0x11040,	// Simple prompt           | OK button               | Default: OK
		MMI_Style_Yes_No1 =        0x11044,	// Simple prompt           | Yes, No buttons         | Default: Yes
		MMI_Style_Yes_No_Cancel1 = 0x11043,	// Simple prompt           | Yes, No, Cancel buttons | Default: Yes
		MMI_Style_Abort_Retry1 =   0x11042, // Simple prompt           | Abort, Retry buttons    | Default: Abort
		MMI_Style_Edit1 =          0x12040, // Request for data input  | OK, Cancel buttons      | Default: OK
		MMI_Style_Edit2 =          0x12140	// Select item from a list | OK, Cancel buttons      | Default: OK
		};

HRESULT CPTSControlClientLogger::OnImplicitSend(LPCWSTR pszProjectName,WORD wID,LPCWSTR pszTestCase,LPCWSTR pszDescription,
	DWORD style,LPWSTR pszResponse,DWORD responseSize,BOOL* pbResponseIsPresent)
{
//
// Select either STANDALONE_TEST or USE_REAL_IMPLICIT_SEND
//
// STANDLONE_TEST: When this function is invoked, display a message box containing a list of all of the parameters to the
//                 function.
//
// USE_REAL_IMPLICIT_SEND: When this function is invoked, invoke PTS's Implicit Send DLL just as if this hook function had
//                         never been invoked. This feature was primarily used in testing this function to make sure that
//                         the parameters sent in would result in the same dialogs that occur when using the normal
//                         Implicit Send DLL.
//
#define	STANDALONE_TEST
//#define	USE_REAL_IMPLICIT_SEND

//
// STANDALONE_TEST
//
#ifdef	STANDALONE_TEST
	wchar_t message[2048];
	swprintf_s(message,_countof(message),
		L"CPTSControlClientLogger::OnImplicitSend() has been called with the following parameters:\r\n"
		L"\r\n"
		L"     pszProjectName = \"%s\"\r\n"
		L"     wID = %d\r\n"
		L"     pszTestCase = \"%s\"\r\n"
		L"     pszDescription = \"%s\"\r\n"
		L"     style = 0x%05x",
		pszProjectName,wID,pszTestCase,pszDescription,style);
	MessageBox(NULL,message,L"CPTSControlClientLogger::OnImplicitSend()",MB_OK);

	DWORD dwResponseSize = responseSize / DWORD(sizeof(wchar_t));

	wcscpy_s(pszResponse, dwResponseSize, L"OK");

	*pbResponseIsPresent = TRUE;
#endif

//
// USE_REAL_IMPLICIT_SEND
//
#ifdef	USE_REAL_IMPLICIT_SEND
	//
	// Interface to the only function that we actually need from the Implicit Send DLL.
	//
	typedef char*	(WINAPI* LPIMPLICITSENDSTYLE)(std::string& strMmiText,UINT mmiStyle);	// "ImplicitSendStyle"
	static LPIMPLICITSENDSTYLE pImplicitSendStyle = NULL;

	//
	// Have we loaded PTS's Implicit Send DLL?
	//
	if (pImplicitSendStyle == NULL)
	{
		// No, do so now
		HMODULE hModule = NULL;	// Yeah I know, I'm going to leak the handle to the library :-)
		const wchar_t* pszImplicitSendDll = L"C:\\Program Files\\Bluetooth SIG\\Bluetooth PTS\\bin\\implicit_send3.dll";
		if ((hModule=LoadLibrary(pszImplicitSendDll)) == NULL)
		{
			swprintf_s(message,_countof(message),L"LoadLibrary(%s) returned %d\n",pszImplicitSendDll,GetLastError());
			Log(PTS_LOGTYPE_ERROR, L"", L"", message);
			return PTSCONTROL_E_IMPLICIT_SEND_CALLBACK_NOT_EXPECTED_TO_FAIL;
		}

		if ((pImplicitSendStyle=(LPIMPLICITSENDSTYLE) GetProcAddress(hModule,"ImplicitSendStyle")) == NULL)
		{
			swprintf_s(message,_countof(message),L"GetProcAddress(ImplicitSendStyle) returned %d\n",GetLastError());
			Log(PTS_LOGTYPE_ERROR, L"", L"", message);
			return PTSCONTROL_E_IMPLICIT_SEND_CALLBACK_NOT_EXPECTED_TO_FAIL;
		}
	}

	//
	// PTS broke the mmi text into pieces before we were called. We need to reassemble it because ImplicitSendStyle() expects
	// a complete message (which it will then break apart...)
	//
	swprintf_s(message,_countof(message),L"{%d,%s,%s}%s",wID,pszTestCase,pszProjectName,pszDescription);
	CStringA csMmiText = CStringA(message);	// Convert from Unicode to MBCS

	//
	// Call the Implicit Send dll
	//
	char* pResponse = pImplicitSendStyle(std::string(csMmiText.GetBuffer()),style);
	csMmiText.ReleaseBuffer();

	//
	// Convert the response into the return parameters
	//
	if (pResponse != NULL)
	{
		CStringW csResponse = CStringW(pResponse);
		wcscpy_s(pszResponse,dwResponseSize,csResponse);
		*pbResponseIsPresent = TRUE;
	}
	else
	{
		*pbResponseIsPresent = FALSE;
	}
#endif

	// Replace the following function call to Log if you want to change the way it is logged.
	HRESULT hr = Log(PTS_LOGTYPE_IMPLICIT_SEND, L"", L"", L"Implicit Send callback OnImplicitSend() has been called.\n");
	return hr;
}


ULONG CPTSControlClientLogger::AddRef()
{
    return InterlockedIncrement(&m_cRef); 
}


ULONG CPTSControlClientLogger::Release()
{
    LONG cRef = InterlockedDecrement(&m_cRef);
    if (0 == cRef)
    {
        delete this;
    }
    return cRef;
}


HRESULT CPTSControlClientLogger::QueryInterface
(
    REFIID      iid,
    void      **ppv
)
{
    if (NULL == ppv)
    {
        return E_POINTER;
    }

    if (__uuidof(IUnknown) == iid || __uuidof(IPTSControlClientLogger) == iid)
    {
        AddRef();
        *ppv = (IPTSControlClientLogger*)this;
        return S_OK;
    }
    else if (__uuidof(IPTSImplicitSendCallback) == iid)
    {
        AddRef();
        *ppv = (IPTSImplicitSendCallback*)this;
        return S_OK;
    }
    else if (__uuidof(IPTSImplicitSendCallbackEx) == iid)
    {
        AddRef();
        *ppv = (IPTSImplicitSendCallbackEx*)this;
        return S_OK;
    }

    *ppv = NULL;
    return E_NOINTERFACE;
}   //   CClassFactory::QueryInterface()

// RAII for CoInitailize
class CCoInitialize
{
public:
    CCoInitialize(HRESULT &hr)
    {
        hr = CoInitializeEx(NULL, COINIT_MULTITHREADED); 
    }

    ~CCoInitialize()
    {
        CoUninitialize();
    }
};

// Global structs and variables
typedef struct TestcaseTag {CString name; BOOL active;} sTestcase, *pTestcase;
typedef struct TestsuiteTag {CString name; std::vector<pTestcase> testcases;} sTestsuite, *pTestsuite;

CComPtr<IPTSControlClientLogger> g_pLog = NULL;

CString g_cTestsuite;
CString g_cName;
CString g_cWorkspace;

CComPtr<IPTSImplicitSendCallback> g_pCustCallback = NULL;
CComPtr<IPTSImplicitSendCallbackEx> g_pCustCallbackEx = NULL;


// Local helper functions used by main
bool clientInitialize(CComPtr<IPTSControl>&);
bool custInitialize(CComPtr<IPTSControl>&);
bool custCleanup(CComPtr<IPTSControl>&);

enum ITAStatus
	{
	TestSuiteNotFound,
	TestCaseNotFound,
	TestCaseFound
	};
ITAStatus isTestcaseAvailable(CString &, CString &, const std::vector<pTestsuite>&);

bool connectToDevice(CComPtr<IPTSControl>&);
void showUsage();
bool clientLoadTestFile(_TCHAR*, std::vector<CString>&);
bool clientLoadWorkspace(CComPtr<IPTSControl>&, std::vector<pTestsuite>&);
bool clientRunTestcases(CComPtr<IPTSControl>&, const std::vector<CString>&, const std::vector<pTestsuite>&);
bool clientShowBdAddress(CComPtr<IPTSControl>&);
void clientCleanup(std::vector<pTestsuite>&);

/* ===========================================================================
 * Function: 	_tmain
 * Description: Main function of PTSControlClient
 * Returns:		int -- 0 if return normally, otherwise 1
 * Arguments:	IN 	int - number of arguments 
 *				IN 	_TCHAR* - array of the arguments
 * =========================================================================== */
int _tmain(int argc, _TCHAR* argv[])
{

	// Check the input parameters
	if( argc != 2 )
	{
		showUsage();
		return 0;
	}

    // Get handles to STDIN and STDOUT. 
    hStdout = GetStdHandle(STD_OUTPUT_HANDLE); 
    if (hStdout == INVALID_HANDLE_VALUE) 
    {
		fwprintf(stderr,L"GetStdHandle(STD_OUTPUT_HANDLE) failed with status %d\n",GetLastError());
		return 1;
    }

    // Save the current text colors. 
    CONSOLE_SCREEN_BUFFER_INFO csbiInfo; 
    if (!GetConsoleScreenBufferInfo(hStdout, &csbiInfo)) 
    {
		fwprintf(stderr,L"GetConsoleScreenBufferInfo() failed with status %d\n",GetLastError());
		return 1;
    }

    wOldColorAttrs = csbiInfo.wAttributes; 

	doLog(LOCAL_MESSAGE, L"", L"", L"Running client...\n");

	HRESULT hr = S_OK;
	
	CCoInitialize ci(hr); 
	if (FAILED(hr))
	{
		wchar_t message[1024];
		swprintf_s(message, _countof(message), L"CoInitialize failed (hr = 0x%x)\n", hr);
		doLog(LOCAL_ERROR, L"", L"", message);
		SetConsoleTextAttribute(hStdout, wOldColorAttrs);
		return 1;   
	}

	CComPtr<IPTSControl> pPTS;

	// initialization procedure
	if( clientInitialize(pPTS) == false )
	{
		SetConsoleTextAttribute(hStdout, wOldColorAttrs);
		return 1;
	}

	// Wait for the PTS app to start
	Sleep(5000);

	// Show Bluetooth Address of PTS
	//clientShowBdAddress(pPTS);

	// Custom initialization procedure
	// The default function will replace the implicit_send call back with CPTSControlClientLogger::OnImplicitSend() or
	// CPTSControlClientLogger::OnSend defined in this file.
	if ( custInitialize(pPTS) == false )
	{
		SetConsoleTextAttribute(hStdout, wOldColorAttrs);
		return 1;
	}

	{
		// Global list of test cases to be tested
		std::vector<CString> vTestcases;
		// test file loading procedure
		if( clientLoadTestFile(argv[1], vTestcases) == false )
		{
			custCleanup(pPTS);
			SetConsoleTextAttribute(hStdout, wOldColorAttrs);
			return 1;
		}

		// List of test suite available in the workspace
		std::vector<pTestsuite> vAllTestsuites;
		// workspace loading procedure
		if( clientLoadWorkspace(pPTS, vAllTestsuites) == false )
		{
			custCleanup(pPTS);
			clientCleanup(vAllTestsuites);
			SetConsoleTextAttribute(hStdout, wOldColorAttrs);
			return 1;
		}

		// run test cases
		if( clientRunTestcases(pPTS, vTestcases, vAllTestsuites) == false)
		{
			custCleanup(pPTS);
			clientCleanup(vAllTestsuites);
			SetConsoleTextAttribute(hStdout, wOldColorAttrs);
			return 1;
		}

		// Custom steps of clean up
		if( custCleanup(pPTS) == false )
		{
			clientCleanup(vAllTestsuites);
			SetConsoleTextAttribute(hStdout, wOldColorAttrs);
			return 1;
		}

		Sleep(3000);

		clientCleanup(vAllTestsuites);
	}

	SetConsoleTextAttribute(hStdout, wOldColorAttrs);

	// Dump memory leaks (only for Debug build)
	_CrtDumpMemoryLeaks();

	return 0;
}


/* ===========================================================================
 * Function: 	clientInitialize
 * Description: Co-create the PTSControl service and setup the logger. This is the default initialization function.
 *				Probably no need to change within this function.
 * Returns:		int -- true if succeeds, otherwise false
 * Arguments:	IN 	CComPtr<IPTSControl> & - the global COM pointer to the PTSControl 
 * =========================================================================== */
bool clientInitialize(CComPtr<IPTSControl> &pPTS)
{

    HRESULT hr = CoCreateInstance(__uuidof(PTSControl), NULL, CLSCTX_ALL, __uuidof(IPTSControl), (void**)&pPTS);
    if (FAILED(hr))
    {
		wchar_t message[1024];
		swprintf_s(message, _countof(message), L"CoCreateInstance(PTSControl) failed (hr = 0x%x)\n", hr);
		doLog(LOCAL_ERROR, L"", L"", message);
        return false;
    }

	g_pLog = new CPTSControlClientLogger(hr);
    if (NULL == g_pLog)
    {
		doLog(LOCAL_ERROR, L"", L"", L"Out of memory allocating CPTSControlClientLogger\n");
        return false;
    }

    if (FAILED(hr))
    {
		wchar_t message[1024];
		swprintf_s(message, _countof(message), L"Failed CPTSControlClientLogger::CPTSControlClientLogger (hr = 0x%x)\n", hr);
		doLog(LOCAL_ERROR, L"", L"", message);
        return false;
    }

	
    hr = pPTS->SetControlClientLoggerCallback(g_pLog);
    if (FAILED(hr))
    {
		wchar_t message[1024];
		swprintf_s(message, _countof(message), L"Failed IPTSControl::SetControlClientLoggerCallback (hr = 0x%x)\n", hr);
		doLog(LOCAL_ERROR, L"", L"", message);
        return false;
    }

	return true;
}

/* ===========================================================================
 * Function: 	custInitialize
 * Description: Additional initialfization to the default process in clientInitialize.
 * Returns:		bool -- true if succeeds, otherwise false
 * Arguments:	IN 	CComPtr<IPTSControl> & - the global COM pointer to the PTSControl 
 * =========================================================================== */
bool custInitialize(CComPtr<IPTSControl> &pPTS)
{
	// Replace with your customization code in this function
	HRESULT hr;
	// Get the callback functions for Implicit Send provided by CPTSControlClientLogger
	// Which Implicit Send callback would the user like to employ?
	switch (MessageBox(NULL,
		L"Do you wish to the use the new Implicit Send callback?\r\n"
		L"\r\n"
		L"Press [Yes] to use the new version (OnImplicitSend()).\r\n"
		L"\r\n"
		L"Press [No] to use the old version (OnSend()).\r\n"
		L"\r\n"
		L"Press [Cancel] to not use either Implicit Send callback.",
		L"PTSControlClient",MB_YESNOCANCEL))
	{
	case IDYES:
		hr = g_pLog->QueryInterface(__uuidof(IPTSImplicitSendCallbackEx), (void**)&g_pCustCallbackEx);
		if (FAILED(hr))
		{
			wchar_t message[1024];
			swprintf_s(message, _countof(message), L"IPTSimplicitSendCallbackEx failed (hr = 0x%x)\n", hr);
			doLog(LOCAL_ERROR, L"", L"", message);
			return false;
		}
		// Register the new one
		hr = pPTS->RegisterImplicitSendCallbackEx(g_pCustCallbackEx);
		if (FAILED(hr))
		{
			wchar_t message[1024];
			swprintf_s(message, _countof(message), L"RegisterImplicitSendCallbackEx failed (hr = 0x%x)\n", hr);
			doLog(LOCAL_ERROR, L"", L"", message);
			return false;
		}
		break;
	case IDNO:
		hr = g_pLog->QueryInterface(__uuidof(IPTSImplicitSendCallback), (void**)&g_pCustCallback);
		if (FAILED(hr))
		{
			wchar_t message[1024];
			swprintf_s(message, _countof(message), L"IPTSimplicitSendCallback failed (hr = 0x%x)\n", hr);
			doLog(LOCAL_ERROR, L"", L"", message);
			return false;
		}
		// Register the old one
		hr = pPTS->RegisterImplicitSendCallback(g_pCustCallback);
		if (FAILED(hr))
		{
			wchar_t message[1024];
			swprintf_s(message, _countof(message), L"RegisterImplicitSendCallback failed (hr = 0x%x)\n", hr);
			doLog(LOCAL_ERROR, L"", L"", message);
			return false;
		}
		break;
	default:
		// Don't register either
		break;
	}

	return true;
}

/* ===========================================================================
 * Function: 	custCleanup
 * Description: Cleanup the customization before exiting PTSControlClient.exe
 *				Probably need to match the customization added in custInitialize
 * Returns:		bool -- true if succeeds, otherwise false
 * Arguments:	IN 	CComPtr<IPTSControl> & - the global COM pointer to the PTSControl 
 * =========================================================================== */
bool custCleanup(CComPtr<IPTSControl> &pPTS)
{
	// Replace with your customization code in this function
	bool bRetVal = true;
	if (g_pCustCallback != NULL)
	{
		HRESULT hr = pPTS->UnregisterImplicitSendCallback(g_pCustCallback);
		if (FAILED(hr))
		{
			bRetVal = false;
		}
	}

	if (g_pCustCallbackEx != NULL)
	{
		HRESULT hr = pPTS->UnregisterImplicitSendCallbackEx(g_pCustCallbackEx);
		if (FAILED(hr))
		{
			bRetVal = false;
		}
	}

	ClosePTS(pPTS);

	if (pPTS != NULL)
	{
		bRetVal = false;
	}

	return bRetVal;
}

/* ===========================================================================
 * Function: 	clientCleanup
 * Description: Deletes previously allocated memory.
 * Arguments:	OUT vector<pTestsuite>& - collection of all testsuits in the workspace
 * =========================================================================== */
void clientCleanup(std::vector<pTestsuite>& vAllTestsuites)
{
	// Clean memory allocated for all testsuites
	while (vAllTestsuites.size() > 0)
	{
		pTestsuite pSuite = vAllTestsuites[vAllTestsuites.size() - 1];
		vAllTestsuites.pop_back();
		if (pSuite != NULL)
		{
			while (pSuite->testcases.size() > 0)
			{
				pTestcase pCase = pSuite->testcases[pSuite->testcases.size() - 1];
				pSuite->testcases.pop_back();
				if (pCase != NULL)
				{
					delete pCase;
				}
			}
			delete pSuite;
		}
	}
}


/* ===========================================================================
 * Function: 	isTestcaseAvailable
 * Description: Go through the global testcase list and return whether the passed in test suite and test case exists or not
 * Returns:		bool - true if the test case exists, otherwise false
 * Arguments:	IN 	CString& - name of the test suite in query
 *				IN 	CString& - name of the test case in query
 *				IN const std::vector<pTestsuite>& - list of all testsuites in the loaded workspace
 * =========================================================================== */

ITAStatus isTestcaseAvailable(CString& runTestsuite, CString& runTestcase, const std::vector<pTestsuite>& vAllTestsuites)
{
	unsigned int iTestsuites = vAllTestsuites.size();
	if (iTestsuites <= 0)
	{
		return TestSuiteNotFound;
	}

	unsigned int i = 0;

	do
	{
		pTestsuite testsuite = vAllTestsuites[i];
		if( testsuite->name.Compare(runTestsuite) == 0 )
		{
			unsigned int iTestcases = testsuite->testcases.size();
			if (iTestcases > 0)
			{
				unsigned int j = 0;
				do
				{
					pTestcase testcase = testsuite->testcases.at(j);
					if( testcase->name.Compare(runTestcase) == 0 && testcase->active)
					{
						return TestCaseFound;
					}
					j++;
				} while(j < iTestcases);

				return TestCaseNotFound;
			}
		}
		i++;
	}while(i < iTestsuites);

	return TestSuiteNotFound;
}

/* ===========================================================================
 * Function: 	showUsage
 * Description: Simply diplay the usage of PTSControlClient.exe
 * Returns:		void
 * Arguments:	void
 * =========================================================================== */
void showUsage()
{

	wprintf(L"PTSControlServer.exe\n");
	wprintf(L"Usage: PTSControlServer [filename]\n");
	wprintf(L"\tfilename -- file name with path information of the automation script.\n");

}

/* ===========================================================================
 * Function: 	clientLoadTestFile
 * Description: Open the testscript file and extract the information including the test cases to be tested.
 *				The testscript should be an XML file and matches the schema TestScriptSchema.xsd.
 * Returns:		bool - true if the test case exists, otherwise false
 * Arguments:	IN 	_TCHAR* - name with path of the testscript file
 *				OUT std::vector<CString>& - list of testcases to run
 * =========================================================================== */
bool clientLoadTestFile(_TCHAR* testFile, std::vector<CString>& vTestcases)
{

	// Read the test file
	FILE* fd = NULL;
	if (_wfopen_s(&fd,testFile, L"r") != 0)
	{
		wchar_t message[1024];
		swprintf_s(message, _countof(message), L"Not able to open the testscript: %s\n", testFile);
		doLog(LOCAL_ERROR, L"", L"", message);
		return false;
	}

	fseek( fd, 0, SEEK_END);
	unsigned int iFileSize = ftell(fd);
	rewind(fd);

	char* scriptXML = (char*) malloc((sizeof(char)*iFileSize) + 1);
	if( scriptXML == NULL )
	{
		doLog(LOCAL_ERROR, L"", L"", L"Not able to allocate memory\n");
		return false;
	}
	memset(scriptXML, 0, iFileSize);

	unsigned int iRead = fread((void*)scriptXML, sizeof(char), iFileSize, fd);


	// Create XML instance
	MSXML2::IXMLDOMDocumentPtr pXMLDom;
	HRESULT hr = pXMLDom.CreateInstance(__uuidof(MSXML2::DOMDocument60));
	if (FAILED(hr))
	{
		doLog(LOCAL_ERROR, L"", L"", L"Failed to create XMLDom instance\n");
		return false;
	}


	if ( pXMLDom->loadXML(scriptXML) != VARIANT_TRUE)
	{
		wchar_t message[1024];
		swprintf_s(message, _countof(message), L"Failed to load data from file %s.\n", testFile);
		doLog(LOCAL_ERROR, L"", L"", message);
		free(scriptXML);
		return false;
	}

	free(scriptXML);

	// Parsing the XML file and read the data
	MSXML2::IXMLDOMNodeListPtr pInfoNodes = pXMLDom->selectNodes("//AUTOMATION");
	if( pInfoNodes->length == 0 )
	{
		doLog(LOCAL_ERROR, L"", L"", L"AUTOMATION node not found\n");
		return false;
	}
	MSXML2::IXMLDOMNodePtr pNode = pInfoNodes->item[0];
	if(pNode){
		MSXML2::IXMLDOMNodePtr pName = pNode->selectSingleNode("Name");
		MSXML2::IXMLDOMNodePtr pWorkspace = pNode->selectSingleNode("Workspace");
		MSXML2::IXMLDOMNodePtr pTestsuite = pNode->selectSingleNode("Testsuite");
		if( pName == NULL || pWorkspace == NULL || pTestsuite == NULL )
		{
			doLog(LOCAL_ERROR, L"", L"", L"Not able to find the requested node in the test script.\n");
			return false;
		}

		wchar_t message[1024];
		g_cName = (char*)pName->text;
		swprintf_s(message, _countof(message), L"Name is %s\n", g_cName.GetBuffer());
		doLog(LOCAL_MESSAGE, L"", L"", message);

		g_cWorkspace = (char*)pWorkspace->text;
		swprintf_s(message, _countof(message), L"Workspace is %s\n", g_cWorkspace.GetBuffer());
		doLog(LOCAL_MESSAGE, L"", L"", message);

		g_cTestsuite = (char*)pTestsuite->text;
		swprintf_s(message, _countof(message), L"Testsuite is %s\n", g_cTestsuite.GetBuffer());
		doLog(LOCAL_MESSAGE, L"", L"", message);

	}
	else
	{
		doLog(LOCAL_MESSAGE, L"", L"", L"Item not found\n");
		return false;
	}

	// Get testcase now
	pInfoNodes = pXMLDom->selectNodes("//AUTOMATION/Testcase");
	if( pInfoNodes->length == 0 )
	{
		doLog(LOCAL_ERROR, L"", L"", L"Testcase not found\n");
		return false;
	}

	unsigned int iNumTestcase = pInfoNodes->length;
	for( unsigned int i=0; i < iNumTestcase; i++ )
	{
		pNode = pInfoNodes->item[i];
		if( pNode == NULL )
		{
			doLog(LOCAL_ERROR, L"", L"", L"Not able to find the requested node in the test script.\n");
			return false;
		}

		
		vTestcases.push_back(CString((char*)pNode->text));
	}

	return true;
}

/* ===========================================================================
 * Function: 	clientLoadWorkspace
 * Description: Through PTSControl, open the workspace specified in the testscript file.
 *				Also, go through all test suites and populate the global test suite list.
 * Returns:		bool - true if the test case exists, otherwise false
 * Arguments:	IN 	CComPtr<IPTSControl> & - the global COM pointer to the PTSControl
 *				OUT	std::vector<pTestsuite>& - list of all testsuites in the loaded workspace
 * =========================================================================== */
bool clientLoadWorkspace(CComPtr<IPTSControl> &pPTS, std::vector<pTestsuite>& vAllTestsuites)
{

	// Load a workspace
    HRESULT hr = pPTS->OpenWorkspace(g_cWorkspace.GetBuffer());
    if (FAILED(hr))
    {
		wchar_t message[1024];
		swprintf_s(message, _countof(message), L"Failed OpenWorkspace (hr = 0x%x)",hr);
		doLog(LOCAL_ERROR, L"", L"", message);
		return false;
    }

	// Retrieve project information and populate local records
    UINT cProjects = 0;
    hr = pPTS->GetProjectCount(&cProjects);
    if (FAILED(hr))
    {
		wchar_t message[1024];
		swprintf_s(message, _countof(message), L"Failed GetProjectCount (hr = 0x%x)",hr);
		doLog(LOCAL_ERROR, L"", L"", message);
		return false;
    }

	DWORD dwPTSVersion = 0;
	hr = pPTS->GetPTSVersion(&dwPTSVersion);
	if( hr == PTSCONTROL_E_PTS_VERSION_NOT_FOUND )
	{
		doLog(LOCAL_ERROR, L"", L"", L"PTS version not found\n");
	}
	else
	{
		wchar_t message[1024];
		swprintf_s(message, _countof(message), L"PTS version is 0x%x\n", dwPTSVersion);
		doLog(LOCAL_MESSAGE, L"", L"", message);
	}

    for (UINT iProject = 0; iProject < cProjects; iProject++)
    {
        CComHeapPtr<WCHAR> pszProjectName;
        hr = pPTS->GetProjectName(iProject, &pszProjectName);
        if (FAILED(hr))
        {
			wchar_t message[1024];
			swprintf_s(message, _countof(message), L"Failed GetProjectNameAtIndex %d (hr = 0x%x)",iProject,hr);
			doLog(LOCAL_ERROR, L"", L"", message);
			return false;
        }

		wchar_t message[1024];
		swprintf_s(message, _countof(message), L"Project: %s\n", pszProjectName);
		doLog(LOCAL_MESSAGE, L"", L"", message);

		DWORD dwProjectVersion = 0;
		hr = pPTS->GetProjectVersion(pszProjectName, &dwProjectVersion);
		if( hr == PTSCONTROL_E_PROJECT_VERSION_NOT_FOUND )
		{
			wchar_t message[1024];
			swprintf_s(message, _countof(message), L"Project %s version not found\n", pszProjectName);
			doLog(LOCAL_ERROR, L"", L"", message);
			return false;
		}
		else
		{
			wchar_t message[1024];
			swprintf_s(message, _countof(message), L"Project %s has version 0x%x\n", pszProjectName, dwProjectVersion);
			doLog(LOCAL_MESSAGE, L"", L"", message);
		}

		pTestsuite newTestsuite = new sTestsuite();
		newTestsuite->name.SetString(pszProjectName);

        UINT cTestCases;
        hr = pPTS->GetTestCaseCount(pszProjectName, &cTestCases);
        if (FAILED(hr))
        {
			wchar_t message[1024];
			swprintf_s(message, _countof(message), L"Failed GetTestCaseCount (hr = 0x%x)\n", hr);
			doLog(LOCAL_ERROR, L"", L"", message);
			return false;
        }


        for (UINT iTestCase = 0; iTestCase < cTestCases; iTestCase++)
        {
            CComHeapPtr<WCHAR> pszTestCaseName;
            hr = pPTS->GetTestCaseName(pszProjectName, iTestCase, &pszTestCaseName);
            if (FAILED(hr))
            {
				wchar_t message[1024];
				swprintf_s(message, _countof(message), L"Failed GetTestCaseName (hr = 0x%x)\n", hr);
				doLog(LOCAL_ERROR, L"", L"", message);
				return false;
            }

            BOOL bIsActive = FALSE;
            HRESULT hr = pPTS->IsActiveTestCase(pszProjectName, pszTestCaseName, &bIsActive);
			if( hr == PTSCONTROL_E_TESTCASE_NOT_FOUND )
			{
				continue;
			}

			if (FAILED(hr))
            {
				wchar_t message[1024];
				swprintf_s(message, _countof(message), L"Failed IsActiveTestCase (hr = 0x%x)\n", hr);
				doLog(LOCAL_ERROR, L"", L"", message);
				return false;
            }

			pTestcase newTestcase = new sTestcase();
			newTestcase->name.SetString(pszTestCaseName);
			newTestcase->active = bIsActive;

			newTestsuite->testcases.push_back(newTestcase);

			wchar_t message[1024];
			swprintf_s(message, _countof(message), L"\tTestCase: %s %s\n", pszTestCaseName, (bIsActive == TRUE)? L"ACTIVE" : L"INACTIVE");
			doLog(LOCAL_MESSAGE, L"", L"", message);
        }

		vAllTestsuites.push_back(newTestsuite);
	}

	return true;

}

/* ===========================================================================
 * Function: 	clientRunTestcase
 * Description: Run test cases that are in the testscript file.
 * Returns:		bool - true if the test case exists, otherwise false
 * Arguments:	IN 	CComPtr<IPTSControl> & - the global COM pointer to the PTSControl
 *				IN	const std::vector<CString>& - list of testcases to run
 *				IN	const std::vector<pTestsuite>& - list of all testsuites in the loaded workspace
 * =========================================================================== */
bool clientRunTestcases(CComPtr<IPTSControl> &pPTS, const std::vector<CString>& vTestcases, const std::vector<pTestsuite>& vAllTestsuites)
{

	unsigned int iNumTestcase = vTestcases.size();
	for( unsigned int i=0; i < iNumTestcase; i++ )
	{
		ITAStatus status;

		CString testcase(vTestcases[i]); 
		if( (status=isTestcaseAvailable(g_cTestsuite, testcase, vAllTestsuites)) == TestCaseFound )
		{
			bool bConnected = connectToDevice(pPTS);
			if (bConnected)
			{
				wchar_t message[1024];
				swprintf_s(message, _countof(message), L"Running TestCase %s\n", vTestcases[i]);
				doLog(PTS_LOGTYPE_START_TEST, L"", L"", message);

				HRESULT hr = pPTS->RunTestCase(g_cTestsuite, vTestcases[i]);

				swprintf_s(message, _countof(message), L"Done running TestCase %s\n", vTestcases[i]);
				doLog(PTS_LOGTYPE_END_TEST, L"", L"", message);

				// Allow device to settle down
				Sleep(3000);
			}
			else
			{
				wchar_t message[1024];
				swprintf_s(message, _countof(message), L"Failed connecting to PTS radio.\n");
				doLog(LOCAL_ERROR, L"", L"", message);
				return false;
			}
		}
		else if (status == TestCaseNotFound)
		{
			wchar_t message[1024];
			swprintf_s(message,_countof(message),L"Test case: %s is not available for testing in loaded testsuite %s.\n", vTestcases[i], g_cTestsuite);
			doLog(LOCAL_ERROR,L"",L"",message);
			return false;
		}
		else
		{
			wchar_t message[1024];
			swprintf_s(message,_countof(message),L"Test suite: %s is not available for testing in loaded workspace %s.\n", g_cTestsuite, g_cWorkspace);
			doLog(LOCAL_ERROR,L"",L"",message);
			return false;
		}
	}

	return true;

}

/* ===========================================================================
* Function: 	connectToDevice
* Description: Retrieve the list of available PTS radio devices and connect to one of them.
* Returns:		bool
* Arguments:	IN 	CComPtr<IPTSControl> & - COM pointer to the PTSControl
* =========================================================================== */
bool connectToDevice(CComPtr<IPTSControl> &pPTS)
{
	bool bRetVal = false;
	if (pPTS.p != NULL)
	{
		CComQIPtr<IPTSControlDeviceSelector> pPTSDevice = pPTS;
		if (pPTSDevice.p != NULL)
		{
			std::vector<std::wstring> vDevices;
			CComVariant vt;
			HRESULT hr = pPTSDevice->GetDeviceList(&vt);
			if (FAILED(hr))
			{
				wchar_t message[1024];
				swprintf_s(message, _countof(message), L"Failed GetDeviceList (hr = 0x%x)\n", hr);
				doLog(LOCAL_ERROR, L"", L"", message);
				return false;
			}
			hr = VariantToStringVector(vt, vDevices);
			if (FAILED(hr))
			{
				wchar_t message[1024];
				swprintf_s(message, _countof(message), L"Failed VariantToStringVector (hr = 0x%x)\n", hr);
				doLog(LOCAL_ERROR, L"", L"", message);
				return false;
			}
			size_t cDevices = vDevices.size();
			for (size_t i = 0; i < cDevices && bRetVal == false; i++)
			{
				if (vDevices[i].compare(0, 10, L"USB:InUse:") != 0)
				{
					std::wstring sDevice = vDevices[i];
					size_t attempt = 0;
					while (bRetVal == false)
					{
						hr = pPTSDevice->SelectDevice(sDevice.c_str());
						if (FAILED(hr))
						{
							wchar_t message[1024];
							swprintf_s(message, _countof(message), L"Failed SelectDevice (hr = 0x%x)\n", hr);
							doLog(LOCAL_ERROR, L"", L"", message);
							return false;
						}
						else
						{
							attempt++;
							if (clientShowBdAddress(pPTS))
							{
								bRetVal = true;
							}
							else
							{
								pPTSDevice->SelectDevice(L"");
								if (attempt >= 2)
								{
									wchar_t message[1024];
									swprintf_s(message, _countof(message), L"Failed connecting to dongle.\n", hr);
									doLog(LOCAL_ERROR, L"", L"", message);
									break;
								}
							}
						}
					}
				}
			}
		}
		else
		{
			wchar_t message[1024];
			swprintf_s(message, _countof(message), L"The installed version of PTSControl does not support IPTSControlDeviceSelector interface.\n");
			doLog(LOCAL_ERROR, L"", L"", message);
			return false;
		}

	}
	return bRetVal;
}

/* ===========================================================================
 * Function: 	clientShowBdAddress
 * Description: Retrieve Bluetooth Address from PTSControl API and display it.
 * Returns:		void
 * Arguments:	IN 	CComPtr<IPTSControl> & - the global COM pointer to the PTSControl
 * =========================================================================== */
bool clientShowBdAddress(CComPtr<IPTSControl> &pPTS)
{
	// Get the bdaddr of the PTS endpoint device
	bool done = false;
	int failCounter = 0;
	HRESULT hr;

	while (!done)
	{
		// We might have to try this more than once because the bdaddr is not available immediately after PTS is started
		if (SUCCEEDED(hr=pPTS->GetPTSBluetoothAddress(&ptsBdAddr)))
		{
			// We got it, display the bdaddr
			wchar_t message[1024];
			swprintf_s(message, _countof(message), L"PTS endpoint address = %012I64X\n",ptsBdAddr);	// Display only the lower 6 bytes
			doLog(LOCAL_MESSAGE, L"", L"", message);
			done = true;
		}
		else
		{
			// Why did the call fail?
			if ((hr == PTSCONTROL_E_BLUETOOTH_ADDRESS_NOT_FOUND)
			 && (++failCounter <= 15))
			{
					// Wait a second and try it again
					doLog(LOCAL_MESSAGE, L"", L"", L"Waiting to get the BDADDR of the PTS endpoint device\n");
					Sleep(1000);
			}
			else
			{
				// Too many tries or some other error
				if (hr == PTSCONTROL_E_BLUETOOTH_ADDRESS_NOT_FOUND)
				{
					doLog(LOCAL_MESSAGE, L"", L"", L"Failed to get the BDADDR of the PTS endpoint device\n");
					return false;
				}
				else
				{
					wchar_t message[1024];
					swprintf_s(message, _countof(message), L"GetPTSBluetoothAddress() failed with status %08x\n",hr);
					doLog(LOCAL_ERROR, L"", L"", message);
					return false;
				}
			}
		}
	}
	return true;
}

/* ===========================================================================
* Function:		ClosePTS
* Description:	Close PTS instance after test execution.
*				This function provided as is for the sample purpose.
* =========================================================================== */
static bool ClosePTS(CComPtr<IPTSControl> &pPTS)
{
    DWORD aProcesses[1024], cbNeeded, cProcesses;
    unsigned int i, j = 0;
	DWORD m_proccessId = GetCurrentProcessId();

	if ( !EnumProcesses( aProcesses, sizeof(aProcesses), &cbNeeded) )
    {
        return false;
    }

    // Calculate how many process identifiers were returned.
    cProcesses = cbNeeded / sizeof(DWORD);

    // check the name and process identifier for each process.
    for ( i = 0; i < cProcesses; i++ )
    {
        if( aProcesses[i] != 0 )
        {
			wchar_t szProcessName[MAX_PATH] = TEXT("<unknown>");
			// Get a handle to the process.
			HANDLE hProcess = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, FALSE, aProcesses[i]);
			// Get the process name.
			if (NULL != hProcess )
			{
				HMODULE hMod;
				DWORD cbNeeded;

				if (EnumProcessModules(hProcess, &hMod, sizeof(hMod), &cbNeeded))
				{
					GetModuleBaseName(hProcess, hMod, szProcessName, sizeof(szProcessName)/sizeof(TCHAR));
				}
			}
			if (_wcsicmp(szProcessName, L"PTS.EXE") == 0)
			{	// Found it. Release the handle to the process.
				HWND hPTS = GetHwndFromProccess(hProcess);
				if(WaitForSingleObject(hProcess, 5000)!=WAIT_OBJECT_0)
				{
					if(hPTS)
					{
						// Release the handle to the process.
						CloseHandle(hProcess);
						pPTS = NULL;
						return true;
					}
				}
			}
			// Release the handle to the process.
			CloseHandle(hProcess);
        }
    }
	return false;
}

/* ===========================================================================
* Function:    GetHwndFromProccess
* Description: Check to see the application is running
* =========================================================================== */
static HWND GetHwndFromProccess(HANDLE hProcess)
{
	getWnd_parameters p_getHwnd;
	p_getHwnd.processId = GetProcessId(hProcess);
	p_getHwnd.hWnd = NULL;

	EnumWindows((WNDENUMPROC)GetProcessWndEnum, (LPARAM) &p_getHwnd);
	return p_getHwnd.hWnd;
}

BOOL CALLBACK GetProcessWndEnum( HWND hwnd, LPARAM lParam )
{
	getWnd_parameters *p_getWnd = (getWnd_parameters *)lParam;
	DWORD dwID;

	GetWindowThreadProcessId(hwnd, &dwID);

	if(dwID == p_getWnd->processId)
	{	// got it
		wchar_t caption[512], s_caption[512];
		memset(caption, 0, 512*sizeof(wchar_t));
		memset(s_caption, 0, 512*sizeof(wchar_t));
		swprintf_s(s_caption, _countof(caption), L"PTS - %012I64X",ptsBdAddr);	// the lower 6 bytes
		GetWindowText(hwnd, caption, 512*sizeof(wchar_t));
		if ((wcsstr(caption, s_caption) != NULL))
		{
			p_getWnd->hWnd = hwnd;
			PostMessage(hwnd, WM_CLOSE, 0, 0) ;
			return FALSE;	// stop it
		}
	}

	return TRUE;	// need to return true to continue
}

